# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to build PHX-Variant from Honeybee Rooms"""

from typing import Dict, Set

from honeybee import room
from honeybee_energy.properties.room import RoomEnergyProperties

from honeybee_ph import site, phi, phius, space
from honeybee_ph.bldg_segment import BldgSegment
from honeybee_ph.properties.room import RoomPhProperties

from honeybee_energy_ph.properties.load import equipment, people
from honeybee_energy_ph.properties.hvac.idealair import IdealAirSystemPhProperties
from honeybee_energy_ph.properties.hot_water.hw_system import SHWSystemPhProperties
from honeybee_energy_ph.hvac.heating import PhHeatingSystem
from honeybee_energy_ph.hvac.cooling import PhCoolingSystem
from honeybee_energy_ph.hvac.ventilation import (
    PhVentilationSystem,
    _ExhaustVentilatorBase,
)
from honeybee_energy_ph.hvac.supportive_device import PhSupportiveDevice
from honeybee_energy_ph.hvac.renewable_devices import PhRenewableEnergyDevice

from PHX.model import phx_site, project, constructions, certification
from PHX.model.utilization_patterns import (
    UtilizationPatternCollection_Ventilation,
    UtilizationPatternCollection_Occupancy,
    UtilizationPatternCollection_Lighting,
)
from PHX.model.enums import (
    phi_certification_phpp_9,
    phi_certification_phpp_10,
    phius_certification,
    hvac,
)
from PHX.from_HBJSON import (
    create_building,
    create_hvac,
    create_elec_equip,
    create_foundations,
)
from PHX.from_HBJSON.create_shw import (
    build_phx_piping,
    build_phx_piping,
    build_phx_hw_heater,
    build_phx_hw_storage,
)


def add_building_from_hb_room(
    _variant: project.PhxVariant,
    _hb_room: room.Room,
    _assembly_dict: Dict[str, constructions.PhxConstructionOpaque],
    _window_type_dict: Dict[str, constructions.PhxConstructionWindow],
    _vent_sched_collection: UtilizationPatternCollection_Ventilation,
    _occ_sched_collection: UtilizationPatternCollection_Occupancy,
    _lighting_sched_collection: UtilizationPatternCollection_Lighting,
    group_components: bool = False,
    _tolerance: float = 0.001,
) -> None:
    """Create the  PHX-Building with all Components and Zones based on a Honeybee-Room.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the building to.
        * _hb_room (room.Room): The honeybee-Room to use as the source.
        * _assembly_dict (Dict[str, constructions.PhxConstructionOpaque]): The Assembly Type dict.
        * _window_type_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.
        * _vent_sched_collection (UtilizationPatternVentCollection): The PhxProject's
            UtilizationPatternVentCollection with the ventilation schedules.
        * _occ_sched_collection (UtilizationPatternOccCollection): The PhxProject's
            UtilizationPatternOccCollection with the occupancy schedules.
        * _lighting_sched_collection (UtilizationPatternLightingCollection): The PhxProject's
            UtilizationPatternLightingCollection with the lighting schedules.
        * group_components (bool): default=False. Set to true to have the converter
            group the components by assembly-type.

    Returns:
    --------
        * None
    """
    _variant.building.add_components(
        create_building.create_components_from_hb_room(
            _hb_room, _assembly_dict, _window_type_dict, _tolerance
        )
    )
    _variant.building.add_zones(
        create_building.create_zones_from_hb_room(
            _hb_room,
            _vent_sched_collection,
            _occ_sched_collection,
            _lighting_sched_collection,
        )
    )

    if group_components:
        _variant.building.merge_opaque_components_by_assembly()
        _variant.building.merge_aperture_components_by_assembly()


def add_phius_certification_from_hb_room(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Set all the PhxPhiusCertificationCriteria on a PhxVariant based on a Honeybee-Room's Building Segment.

    Arguments:
    ----------
        * _variant (project.Variant): The PhxVariant to set the PhxPhiusCertificationCriteria on.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    # -- alias cus' all this shit is deep in there...
    hbph_phius_cert: phius.PhiusCertification = _hb_room.properties.ph.ph_bldg_segment.phius_certification  # type: ignore # alias
    phx_phius_cert_criteria = _variant.phius_cert.phius_certification_criteria  # alias
    phx_phius_cert_settings = _variant.phius_cert.phius_certification_settings  # alias

    # -- some random bullshit
    phx_phius_cert_criteria.ph_selection_target_data = (
        hbph_phius_cert.localization_selection_type
    )

    # -- certification thresholds (for Phius, they change with climate)
    phx_phius_cert_criteria.phius_annual_heating_demand = (
        hbph_phius_cert.PHIUS2021_heating_demand
    )
    phx_phius_cert_criteria.phius_annual_cooling_demand = (
        hbph_phius_cert.PHIUS2021_cooling_demand
    )
    phx_phius_cert_criteria.phius_peak_heating_load = (
        hbph_phius_cert.PHIUS2021_heating_load
    )
    phx_phius_cert_criteria.phius_peak_cooling_load = (
        hbph_phius_cert.PHIUS2021_cooling_load
    )

    # -- certification settings / types
    phx_phius_cert_settings.phius_building_certification_program = (
        phius_certification.PhiusCertificationProgram(
            hbph_phius_cert.certification_program.number
        )
    )
    phx_phius_cert_settings.phius_building_category_type = (
        phius_certification.PhiusCertificationBuildingCategoryType(
            hbph_phius_cert.building_category_type.number
        )
    )
    phx_phius_cert_settings.phius_building_use_type = (
        phius_certification.PhiusCertificationBuildingUseType(
            hbph_phius_cert.building_use_type.number
        )
    )
    phx_phius_cert_settings.phius_building_status = (
        phius_certification.PhiusCertificationBuildingStatus(
            hbph_phius_cert.building_status.number
        )
    )
    phx_phius_cert_settings.phius_building_type = (
        phius_certification.PhiusCertificationBuildingType(
            hbph_phius_cert.building_type.number
        )
    )

    return None


def set_phx_phpp9_settings(
    _phx_settings: certification.PhxPhiCertificationSettings,
    _attributes: phi.PHPPSettings9,
) -> None:
    """Set the values of the PHX-PHI Settings object for PHPP-v9 Settings.

    Arguments:
    ----------
        * _phx_settings: certification.PhxPhiCertificationSettings
        * _hbph_settings:phi.PhiCertification)

    Returns:
    --------
        * None
    """
    _phx_settings.phi_building_category_type = (
        phi_certification_phpp_9.PhiCertBuildingCategoryType(
            _attributes.building_category_type.number
        )
    )
    _phx_settings.phi_building_use_type = phi_certification_phpp_9.PhiCertBuildingUseType(
        _attributes.building_use_type.number
    )
    _phx_settings.phi_building_ihg_type = phi_certification_phpp_9.PhiCertIHGType(
        _attributes.ihg_type.number
    )
    _phx_settings.phi_building_occupancy_type = (
        phi_certification_phpp_9.PhiCertOccupancyType(_attributes.occupancy_type.number)
    )

    _phx_settings.phi_certification_type = phi_certification_phpp_9.PhiCertType(
        _attributes.certification_type.number
    )
    _phx_settings.phi_certification_class = phi_certification_phpp_9.PhiCertClass(
        _attributes.certification_class.number
    )
    _phx_settings.phi_pe_type = phi_certification_phpp_9.PhiCertificationPEType(
        _attributes.primary_energy_type.number
    )
    _phx_settings.phi_enerphit_type = phi_certification_phpp_9.PhiCertEnerPHitType(
        _attributes.enerphit_type.number
    )
    _phx_settings.phi_retrofit_type = phi_certification_phpp_9.PhiCertRetrofitType(
        _attributes.retrofit_type.number
    )


def set_phx_phpp10_settings(
    _phx_settings: certification.PhxPhiCertificationSettings,
    _attributes: phi.PHPPSettings10,
) -> None:
    """Set the values of the PHX-PHI Settings object for PHPP-v10 Settings.

    Arguments:
    ----------
        * _phx_settings: certification.PhxPhiCertificationSettings
        * _hbph_settings:phi.PhiCertification)

    Returns:
    --------
        * None
    """
    _phx_settings.phi_building_use_type = (
        phi_certification_phpp_10.PhiCertBuildingUseType(
            _attributes.building_use_type.number
        )
    )
    _phx_settings.phi_building_ihg_type = phi_certification_phpp_10.PhiCertIHGType(
        _attributes.ihg_type.number
    )

    _phx_settings.phi_certification_type = phi_certification_phpp_10.PhiCertType(
        _attributes.certification_type.number
    )
    _phx_settings.phi_certification_class = phi_certification_phpp_10.PhiCertClass(
        _attributes.certification_class.number
    )
    _phx_settings.phi_pe_type = phi_certification_phpp_10.PhiCertificationPEType(
        _attributes.primary_energy_type.number
    )
    _phx_settings.phi_retrofit_type = phi_certification_phpp_10.PhiCertRetrofitType(
        _attributes.retrofit_type.number
    )


def add_phi_certification_from_hb_room(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Set all the PhxPhiCertificationCriteria on a PhxVariant based on a Honeybee-Room's Building Segment.

    Arguments:
    ----------
        * _variant (project.Variant): The PhxVariant to set the PhxPhiCertificationCriteria to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """
    # alias cus' all this shit is deep in there...
    hbph_settings: phi.PhiCertification = _hb_room.properties.ph.ph_bldg_segment.phi_certification  # type: ignore
    phx_phi_cert = _variant.phi_cert
    phx_settings = _variant.phi_cert.phi_certification_settings  # type: ignore

    if isinstance(hbph_settings.attributes, phi.PHPPSettings10):
        phx_phi_cert.version = 10
        set_phx_phpp10_settings(phx_settings, hbph_settings.attributes)
    elif isinstance(hbph_settings.attributes, phi.PHPPSettings9):
        phx_phi_cert.version = 9
        set_phx_phpp9_settings(phx_settings, hbph_settings.attributes)
    else:
        msg = (
            "Error: Unknown PHPP Settings Version? Expected 9 | 10, "
            f"Got: '{hbph_settings.attributes.phpp_version}'"
        )
        raise Exception(msg)

    return None


def add_PhxPhBuildingData_from_hb_room(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Create and add a PhxPhBuildingData with data from a Honeybee-Room to a PHX-Variant.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the PH-Building to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    # -- Type Aliases
    ph_bldg = _variant.phius_cert.ph_building_data  # alias
    hb_prop_energy: RoomEnergyProperties = _hb_room.properties.energy  # type: ignore
    hb_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    hbph_bldg_seg: BldgSegment = hb_prop_ph.ph_bldg_segment

    # -- Set the occupancy
    if hb_prop_energy.people:
        hb_ppl_prop_ph: people.PeoplePhProperties = hb_prop_energy.people.properties.ph  # type: ignore
        ph_bldg.num_of_units = hb_ppl_prop_ph.number_dwelling_units
    ph_bldg.num_of_floors = hbph_bldg_seg.num_floor_levels  # type: ignore

    # -- Add Foundations
    for hbph_foundation in hb_prop_ph.ph_foundations:
        phx_foundation = create_foundations.create_phx_foundation_from_hbph(
            hbph_foundation
        )
        ph_bldg.add_foundation(phx_foundation)

    # -- Set the airtightness for Building
    ph_bldg.airtightness_q50 = hb_prop_energy.infiltration.flow_per_exterior_area * 3600

    # -- Set the air temp setpoints
    ph_bldg.setpoints.winter = hbph_bldg_seg.set_points.winter
    ph_bldg.setpoints.summer = hbph_bldg_seg.set_points.summer
    ph_bldg.mech_room_temp = hbph_bldg_seg.mech_room_temp
    ph_bldg.non_combustible_materials = hbph_bldg_seg.non_combustible_materials

    # -- Summer HRV Bypass mode
    _ = hbph_bldg_seg.summer_hrv_bypass_mode.number
    ph_bldg.summer_hrv_bypass_mode = hvac.PhxSummerBypassMode(_)

    return None


def add_climate_from_hb_room(_variant: project.PhxVariant, _hb_room: room.Room) -> None:
    """Copy PHX-Climate info from a Honeybee-Room over to a PHX-Variant.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the climate data to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    # -- aliases
    phx_climate = _variant.site.climate
    ud_site: site.Site = _hb_room.properties.ph.ph_bldg_segment.site  # type: ignore
    ud_ground = ud_site.climate.ground

    # -- Location
    _variant.site.location.latitude = ud_site.location.latitude
    _variant.site.location.longitude = ud_site.location.longitude
    _variant.site.location.site_elevation = ud_site.location.site_elevation
    _variant.site.location.climate_zone = ud_site.location.climate_zone
    _variant.site.location.hours_from_UTC = ud_site.location.hours_from_UTC

    # -- Basics
    phx_climate.station_elevation = ud_site.climate.station_elevation
    phx_climate.daily_temp_swing = ud_site.climate.summer_daily_temperature_swing
    phx_climate.avg_wind_speed = ud_site.climate.average_wind_speed

    # -- PHPP Stuff
    _variant.site.phpp_codes.country_code = ud_site.phpp_library_codes.country_code
    _variant.site.phpp_codes.region_code = ud_site.phpp_library_codes.region_code
    _variant.site.phpp_codes.dataset_name = ud_site.phpp_library_codes.dataset_name

    # -- Ground
    _variant.site.ground.ground_thermal_conductivity = (
        ud_ground.ground_thermal_conductivity
    )
    _variant.site.ground.ground_heat_capacity = ud_ground.ground_heat_capacity
    _variant.site.ground.ground_density = ud_ground.ground_density
    _variant.site.ground.depth_groundwater = ud_ground.depth_groundwater
    _variant.site.ground.flow_rate_groundwater = ud_ground.flow_rate_groundwater

    # -- Monthly Values
    phx_climate.temperature_air = ud_site.climate.monthly_temps.air_temps.values
    phx_climate.temperature_dewpoint = ud_site.climate.monthly_temps.dewpoints.values
    phx_climate.temperature_sky = ud_site.climate.monthly_temps.sky_temps.values

    phx_climate.radiation_north = ud_site.climate.monthly_radiation.north.values
    phx_climate.radiation_east = ud_site.climate.monthly_radiation.east.values
    phx_climate.radiation_south = ud_site.climate.monthly_radiation.south.values
    phx_climate.radiation_west = ud_site.climate.monthly_radiation.west.values
    phx_climate.radiation_global = ud_site.climate.monthly_radiation.glob.values

    # -- Peak Load Values
    phx_climate.peak_heating_1.temperature_air = (
        ud_site.climate.peak_loads.heat_load_1.temp
    )
    phx_climate.peak_heating_1.radiation_north = (
        ud_site.climate.peak_loads.heat_load_1.rad_north
    )
    phx_climate.peak_heating_1.radiation_east = (
        ud_site.climate.peak_loads.heat_load_1.rad_east
    )
    phx_climate.peak_heating_1.radiation_south = (
        ud_site.climate.peak_loads.heat_load_1.rad_south
    )
    phx_climate.peak_heating_1.radiation_west = (
        ud_site.climate.peak_loads.heat_load_1.rad_west
    )
    phx_climate.peak_heating_1.radiation_global = (
        ud_site.climate.peak_loads.heat_load_1.rad_global
    )

    phx_climate.peak_heating_2.temperature_air = (
        ud_site.climate.peak_loads.heat_load_2.temp
    )
    phx_climate.peak_heating_2.radiation_north = (
        ud_site.climate.peak_loads.heat_load_2.rad_north
    )
    phx_climate.peak_heating_2.radiation_east = (
        ud_site.climate.peak_loads.heat_load_2.rad_east
    )
    phx_climate.peak_heating_2.radiation_south = (
        ud_site.climate.peak_loads.heat_load_2.rad_south
    )
    phx_climate.peak_heating_2.radiation_west = (
        ud_site.climate.peak_loads.heat_load_2.rad_west
    )
    phx_climate.peak_heating_2.radiation_global = (
        ud_site.climate.peak_loads.heat_load_2.rad_global
    )

    phx_climate.peak_cooling_1.temperature_air = (
        ud_site.climate.peak_loads.cooling_load_1.temp
    )
    phx_climate.peak_cooling_1.radiation_north = (
        ud_site.climate.peak_loads.cooling_load_1.rad_north
    )
    phx_climate.peak_cooling_1.radiation_east = (
        ud_site.climate.peak_loads.cooling_load_1.rad_east
    )
    phx_climate.peak_cooling_1.radiation_south = (
        ud_site.climate.peak_loads.cooling_load_1.rad_south
    )
    phx_climate.peak_cooling_1.radiation_west = (
        ud_site.climate.peak_loads.cooling_load_1.rad_west
    )
    phx_climate.peak_cooling_1.radiation_global = (
        ud_site.climate.peak_loads.cooling_load_1.rad_global
    )

    phx_climate.peak_cooling_2.temperature_air = (
        ud_site.climate.peak_loads.cooling_load_2.temp
    )
    phx_climate.peak_cooling_2.radiation_north = (
        ud_site.climate.peak_loads.cooling_load_2.rad_north
    )
    phx_climate.peak_cooling_2.radiation_east = (
        ud_site.climate.peak_loads.cooling_load_2.rad_east
    )
    phx_climate.peak_cooling_2.radiation_south = (
        ud_site.climate.peak_loads.cooling_load_2.rad_south
    )
    phx_climate.peak_cooling_2.radiation_west = (
        ud_site.climate.peak_loads.cooling_load_2.rad_west
    )
    phx_climate.peak_cooling_2.radiation_global = (
        ud_site.climate.peak_loads.cooling_load_2.rad_global
    )

    return None


def add_local_pe_conversion_factors(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Copy local Site->Source conversion factors from a Honeybee-Room over to a PHX-Variant.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the factor data to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """
    for factor in _hb_room.properties.ph.ph_bldg_segment.source_energy_factors:  # type: ignore
        new_phx_factor = phx_site.PhxPEFactor()
        new_phx_factor.fuel_name = factor.fuel_name
        new_phx_factor.value = factor.value
        new_phx_factor.unit = factor.unit
        _variant.site.energy_factors.pe_factors[new_phx_factor.fuel_name] = new_phx_factor
    return


def add_local_co2_conversion_factors(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Copy local Site->CO2e conversion factors from a Honeybee-Room over to a PHX-Variant.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the factor data to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    for factor in _hb_room.properties.ph.ph_bldg_segment.co2e_factors:  # type: ignore
        new_phx_factor = phx_site.PhxCO2Factor()
        new_phx_factor.fuel_name = factor.fuel_name
        new_phx_factor.value = factor.value
        new_phx_factor.unit = factor.unit
        _variant.site.energy_factors.co2_factors[
            new_phx_factor.fuel_name
        ] = new_phx_factor
    return


def add_ventilation_systems_from_hb_rooms(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Add new HVAC (Ventilation, etc) Systems to the Variant based on the HB-Rooms.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the new hvac systems to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    ph_prop: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    for hbph_space in ph_prop.spaces:
        # ---------------------------------------------------------------------
        # -- Get the Honeybee-PH Ventilation system from the space's host room
        # -- Note: in the case of a merged room, the space host may not be the same
        # -- as _hb_room, so always refer back to the space.host to be sure.
        hbph_host_room = hbph_space.host
        prop_e: RoomEnergyProperties = hbph_host_room.properties.energy  # type: ignore
        hbph_vent_sys: PhVentilationSystem = prop_e.hvac.properties.ph.ventilation_system  # type: ignore

        if not hbph_vent_sys:
            continue

        # ---------------------------------------------------------------------
        # -- Get or Build the PHX Ventilation Device
        # -- If the ventilator already exists, just use that one.
        phx_ventilator = _variant.mech_systems.get_mech_device_by_key(hbph_vent_sys.key)
        if not phx_ventilator:
            # -- otherwise, build a new PH-Ventilator from the HB-hvac
            phx_ventilator = create_hvac.build_phx_ventilator(hbph_vent_sys)
            _variant.mech_systems.add_new_mech_device(hbph_vent_sys.key, phx_ventilator)

        # ---------------------------------------------------------------------
        # -- Re-set the HBPH Ventilator and sys id_num to keep everything aligned
        hbph_vent_sys.id_num = phx_ventilator.id_num
        if hbph_vent_sys.ventilation_unit:
            hbph_vent_sys.ventilation_unit.id_num = phx_ventilator.id_num

        # ---------------------------------------------------------------------
        # -- Add PHX Distribution Ducting from the HBPH Ducts
        for hbph_supply_duct in hbph_vent_sys.supply_ducting:
            _variant.mech_systems.add_vent_ducting(
                create_hvac.build_phx_duct(hbph_supply_duct, phx_ventilator.id_num)
            )
        for hbph_exhaust_duct in hbph_vent_sys.exhaust_ducting:
            _variant.mech_systems.add_vent_ducting(
                create_hvac.build_phx_duct(hbph_exhaust_duct, phx_ventilator.id_num)
            )

    return None


def add_exhaust_vent_devices_from_hb_rooms(
    _variant: project.PhxVariant,
    _hb_room: room.Room,
    _merge_devices: bool = True,
) -> None:
    """Add all the Exhaust Ventilation Equipment from the HB Room onto the new Variant/Zone.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the new exhaust equipment to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.
        * _merge_devices (bool): Default=True. Merge together the Exhaust Ventilation
            Devices into a single device (per type)?

    Returns:
    --------
        * None
    """

    def _get_space_exhaust_vent_devices(
        _hph_space: space.Space,
    ) -> Set[_ExhaustVentilatorBase]:
        """Return a set of all the ExhaustVentilators found on a space's host HB Room."""
        # -- Get the Honeybee-PH Exhaust Vent Devices from the space's host room
        # -- Note: in the case of a merged room, the space host may not be the same
        # -- as _hb_room, so always refer back to the space.host to be sure.
        host_rm_prop_energy: RoomEnergyProperties = _hph_space.host.properties.energy  # type: ignore
        hvac_prop_ph: IdealAirSystemPhProperties = host_rm_prop_energy.hvac.properties.ph  # type: ignore
        return hvac_prop_ph.exhaust_vent_devices

    room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    for hbph_space in room_prop_ph.spaces:
        for hbph_device in _get_space_exhaust_vent_devices(hbph_space):
            key = hbph_device.key

            # -- Get or Build the PHX Exhaust Ventilation Device
            # -- If the Device already exists, just use that one.
            for zone in _variant.zones:
                # -- If the device is already in the Zone's collection, dont add another one.
                if zone.exhaust_ventilator_collection.device_in_collection(key):
                    continue

                # -- Otherwise, build a new PHX-Exhaust Ventilator from the HBPH-Object
                phx_device = create_hvac.build_phx_exhaust_vent_device(hbph_device)

                #  -- And add the new ventilator to the Zone's set
                zone.exhaust_ventilator_collection.add_new_ventilator(key, phx_device)

    # -- Once all the Exhaust Ventilators have been added to the Zones,
    # -- merge them together
    for zone in _variant.zones:
        if _merge_devices:
            zone.exhaust_ventilator_collection.merge_all_devices()

    return None


def add_heating_systems_from_hb_rooms(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Add a new PHX-Heating SubSystem to the Variant based on the HB-Rooms.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the new hvac systems to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    for space in room_prop_ph.spaces:
        # -- Get the Honeybee-PH Heating Systems from the space's host room
        # -- Note: in the case of a merged room, the space host may not be the same
        # -- as _hb_room, so always refer back to the space.host to be sure.
        host_prop_energy: RoomEnergyProperties = space.host.properties.energy  # type: ignore
        hvac_prop_ph: IdealAirSystemPhProperties = host_prop_energy.hvac.properties.ph  # type: ignore
        heating_systems: Set[PhHeatingSystem] = hvac_prop_ph.heating_systems

        # -- Get or Build the PHX Heating systems
        for hbph_sys in heating_systems:
            # -- If the system already exists, just use that one.
            phx_heating_device = _variant.mech_systems.get_mech_device_by_key(
                hbph_sys.key
            )

            # -- otherwise, build a new PHX-Heating-Sys from the HB-hvac
            if not phx_heating_device:
                phx_heating_device = create_hvac.build_phx_heating_sys(hbph_sys)
                _variant.mech_systems.add_new_mech_device(
                    hbph_sys.key, phx_heating_device
                )

            # -- Keep the ID-Numbers aligned
            setattr(hbph_sys, "id_num", phx_heating_device.id_num)

    return None


def add_cooling_systems_from_hb_rooms(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Add new PHX-Cooling SubSystem to the Variant based on the HB-Rooms.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX-Variant to add the new hvac systems to.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.

    Returns:
    --------
        * None
    """

    room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    for space in room_prop_ph.spaces:
        # -- Get the Honeybee-PH Cooling-Systems from the space's host room
        # -- Note: in the case of a merged room, the space host may not be the same
        # -- as _hb_room, so always refer back to the space.host to be sure.
        room_prop_energy: RoomEnergyProperties = space.host.properties.energy  # type: ignore
        hvac_prop_ph: IdealAirSystemPhProperties = room_prop_energy.hvac.properties.ph  # type: ignore
        cooling_systems: Set[PhCoolingSystem] = hvac_prop_ph.cooling_systems

        # -- Get or Build the PHX-Cooling systems
        for hbph_sys in cooling_systems:
            # -- If the system already exists, just use that one.
            phx_cooling_device = _variant.mech_systems.get_mech_device_by_key(
                hbph_sys.key
            )

            # -- otherwise, build a new PHX-Cooling-Sys from the HB-hvac
            if not phx_cooling_device:
                phx_cooling_device = create_hvac.build_phx_cooling_sys(hbph_sys)
                _variant.mech_systems.add_new_mech_device(
                    hbph_sys.key, phx_cooling_device
                )

            # -- Keep the ID-Numbers aligned
            setattr(hbph_sys, "id_num", phx_cooling_device.id_num)

    return None


def add_dhw_storage_from_hb_rooms(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Add new Service Hot Water Equipment to the Variant based on the HB-Rooms.

    Arguments:
    ----------
        * _variant (project.Variant): The PHX Variant to add the PHX DHW System to.
        * _hb_room (room.Room): The Honeybee room to get the DHW System data from.

    Returns:
    --------
        * None
    """

    room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    for space in room_prop_ph.spaces:
        host_prop_energy: RoomEnergyProperties = space.host.properties.energy  # type: ignore

        if not host_prop_energy.shw or not host_prop_energy.service_hot_water:
            continue

        # -- Build the HW-Tank
        shw_prop_ph: SHWSystemPhProperties = host_prop_energy.shw.properties.ph  # type: ignore
        for hw_tank in shw_prop_ph.tanks:
            if not hw_tank:
                continue

            if _variant.mech_systems.device_in_collection(hw_tank.key):
                continue

            # -- Build a new PHS-HW-Tank from the HB-hvac
            phx_dhw_tank = build_phx_hw_storage(hw_tank)
            _variant.mech_systems.add_new_mech_device(hw_tank.key, phx_dhw_tank)

    return None


def add_dhw_heaters_from_hb_rooms(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """

    Arguments:
    ----------
        *_variant (project.PhxVariant): The PHX Variant to add the PHX DHW Heaters to.
        _hb_room (room.Room): The Honeybee room to get the DHW Heater data from.

    Returns:
    --------
        * None
    """

    room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    for space in room_prop_ph.spaces:
        host_prop_energy: RoomEnergyProperties = space.host.properties.energy  # type: ignore
        if not host_prop_energy.shw:
            continue

        shw_prop_ph: SHWSystemPhProperties = host_prop_energy.shw.properties.ph  # type: ignore
        for heater in shw_prop_ph.heaters:
            if _variant.mech_systems.device_in_collection(heater.identifier):
                continue

            # -- Build a new PHX-HW-Heater from the Honeybee-PH HW-Heater
            phx_hw_heater = build_phx_hw_heater(heater)
            _variant.mech_systems.add_new_mech_device(heater.identifier, phx_hw_heater)


def dhw_recirc_temp(_recirc_temps: Set[float]):
    """Get the DHW recirculation temperature."""
    if len(_recirc_temps) == 0:
        return 60.0
    elif len(_recirc_temps) == 1:
        return _recirc_temps.pop()
    else:
        print(f"Warning: Multiple recirculation temperatures found. {_recirc_temps}")
        return 60.0


def dhw_recirc_hours(_recirc_temps: Set[int]) -> int:
    """Get the number of hours that the DHW recirculation is active."""
    if len(_recirc_temps) == 0:
        return 24
    elif len(_recirc_temps) == 1:
        return _recirc_temps.pop()
    else:
        print(f"Warning: Multiple recirculation run-times found. {_recirc_temps}")
        return 24


def add_dhw_piping_from_hb_rooms(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    ph_prop: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    phx_mech_sys = _variant.mech_systems
    recirc_temps: Set[float] = set()
    recirc_hours: Set[int] = set()

    for space in ph_prop.spaces:
        # -- Get the HB SHW system from the space's host room
        prop_energy: RoomEnergyProperties = space.host.properties.energy  # type: ignore
        if not prop_energy.shw:
            continue
        hb_shw_prop_ph: SHWSystemPhProperties = prop_energy.shw.properties.ph  # type: ignore

        # -- Add the DHW Branch Piping
        for branch_piping_element in hb_shw_prop_ph.branch_piping:
            phx_mech_sys.add_branch_piping(build_phx_piping(branch_piping_element))

        # -- Set the tap points
        phx_mech_sys._distribution_num_hw_tap_points = hb_shw_prop_ph.number_tap_points

        # -- Add the DHW Recirculation Piping
        for recirc_piping_element in hb_shw_prop_ph.recirc_piping:
            phx_mech_sys.add_recirc_piping(build_phx_piping(recirc_piping_element))

        # -- Get the DHW recirculation parameters
        recirc_temps.add(hb_shw_prop_ph.recirc_temp)
        recirc_hours.add(hb_shw_prop_ph.recirc_hours)

    # -- Set the Recirc parameters
    params = phx_mech_sys._distribution_hw_recirculation_params
    params.water_temp = dhw_recirc_temp(recirc_temps)
    params.daily_recirc_hours = dhw_recirc_hours(recirc_hours)

    return None


def add_elec_equip_from_hb_room(
    _variant: project.PhxVariant, _hb_room: room.Room
) -> None:
    """Creates new PHX-Elec-Equipment (Appliances) and adds them to each of the Variant.building.zones

    Arguments:
    ----------
        * _variant (project.Variant): The Variant to add the new elec-equipment to.
        * _hb_room (room.Room): The Honeybee Room to get the elec-equipment from.

    Returns:
    --------
        * None
    """

    ee_properties_ph: equipment.ElectricEquipmentPhProperties = _hb_room.properties.energy.electric_equipment.properties.ph  # type: ignore
    for equip_key, device in ee_properties_ph.equipment_collection.items():
        phx_elec_device = create_elec_equip.build_phx_elec_device(device)
        for zone in _variant.building.zones:
            zone.elec_equipment_collection.add_new_device(equip_key, phx_elec_device)

    return


def add_supportive_devices_from_hb_room(
    _variant: project.PhxVariant,
    _hb_room: room.Room,
    _merge_devices: bool = True,
) -> None:
    def _get_space_supportive_devices(
        _hph_space: space.Space,
    ) -> Set[PhSupportiveDevice]:
        """Return a set of all the SupportiveDevices found on a space's host HB Room."""
        # -- Get the Honeybee-PH Devices from the space's host room
        # -- Note: in the case of a merged room, the space host may not be the same
        # -- as _hb_room, so always refer back to the space.host to be sure.
        host_rm_prop_energy: RoomEnergyProperties = _hph_space.host.properties.energy  # type: ignore
        hvac_prop_ph: IdealAirSystemPhProperties = host_rm_prop_energy.hvac.properties.ph  # type: ignore
        return hvac_prop_ph.supportive_devices  # type: ignore

    room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    for hbph_space in room_prop_ph.spaces:
        for hbph_device in _get_space_supportive_devices(hbph_space):
            # -- Get or Build the PHX Supportive Device
            # -- If the Device already exists, just use that one.
            # -- If the device is already in the Mech System's collection, dont add another one.
            if _variant.mech_systems.supportive_devices.device_in_collection(
                hbph_device.key
            ):
                continue

            # -- Otherwise, build a new PHX-Exhaust Ventilator from the HBPH-Object
            phx_device = create_hvac.build_phx_supportive_device(hbph_device)
            _variant.mech_systems.supportive_devices.add_new_device(
                hbph_device.key, phx_device
            )

    # -- Once all the Supportive Devices have been added to the Zones, merge them together
    if _merge_devices:
        _variant.mech_systems.supportive_devices.merge_all_devices()

    return None


def add_renewable_devices_from_hb_room(
    _variant: project.PhxVariant,
    _hb_room: room.Room,
    _merge_devices: bool = True,
) -> None:
    def _get_space_renewable_devices(
        _hph_space: space.Space,
    ) -> Set[PhRenewableEnergyDevice]:
        """Return a set of all the Renewable Energy Devices found on a space's host HB Room.

        Note: in the case of a merged room, the space host may not be the same
        as _hb_room, so always refer back to the space.host to be sure.
        """
        host_rm_prop_energy: RoomEnergyProperties = _hph_space.host.properties.energy  # type: ignore
        hvac_prop_ph: IdealAirSystemPhProperties = host_rm_prop_energy.hvac.properties.ph  # type: ignore
        return hvac_prop_ph.renewable_devices  # type: ignore

    room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    for hbph_space in room_prop_ph.spaces:
        for hbph_device in _get_space_renewable_devices(hbph_space):
            # -- Get or Build the PHX Renewable Device
            # -- If the Device already exists, just use that one.
            # -- If the device is already in the Mech System's collection, dont add another one.
            if _variant.mech_systems.renewable_devices.device_in_collection(
                hbph_device.key
            ):
                continue

            # -- Otherwise, build a new PHX-Renewable Device from the HBPH-Object
            phx_device = create_hvac.build_phx_renewable_device(hbph_device)
            _variant.mech_systems.renewable_devices.add_new_device(
                hbph_device.key, phx_device
            )

    # -- Once all the Renewable Devices have been added to the Zones, merge them together
    if _merge_devices:
        _variant.mech_systems.renewable_devices.merge_all_devices()

    return None


def from_hb_room(
    _hb_room: room.Room,
    _assembly_dict: Dict[str, constructions.PhxConstructionOpaque],
    _window_type_dict: Dict[str, constructions.PhxConstructionWindow],
    _vent_sched_collection: UtilizationPatternCollection_Ventilation,
    _occ_sched_collection: UtilizationPatternCollection_Occupancy,
    _lighting_sched_collection: UtilizationPatternCollection_Lighting,
    group_components: bool = False,
    _tolerance: float = 0.001,
) -> project.PhxVariant:
    """Create a new PHX-Variant based on a single PH/Honeybee Room.

    Arguments:
    ----------
        * _hb_room (honeybee.room.Room): The honeybee room to base the Variant on.
        * _assembly_dict (Dict[str, constructions.PhxConstructionOpaque]): The Assembly Type dict.
        * _window_type_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.
        * group_components (bool): default=False. Set to true to have the converter
            group the components by assembly-type.

    Returns:
    --------
        * A new Variant object.
    """

    new_variant = project.PhxVariant()

    # -- Keep all the ID numbers aligned
    new_variant.id_num = project.PhxVariant._count
    _hb_room.properties.ph.id_num = new_variant.id_num  # type: ignore
    new_variant.name = _hb_room.display_name

    # -- Build the Variant Elements (Dev. note: order matters!!)
    add_ventilation_systems_from_hb_rooms(new_variant, _hb_room)
    add_heating_systems_from_hb_rooms(new_variant, _hb_room)
    add_cooling_systems_from_hb_rooms(new_variant, _hb_room)
    add_dhw_heaters_from_hb_rooms(new_variant, _hb_room)
    add_dhw_piping_from_hb_rooms(new_variant, _hb_room)
    add_dhw_storage_from_hb_rooms(new_variant, _hb_room)
    add_building_from_hb_room(
        _variant=new_variant,
        _hb_room=_hb_room,
        _assembly_dict=_assembly_dict,
        _window_type_dict=_window_type_dict,
        _vent_sched_collection=_vent_sched_collection,
        _occ_sched_collection=_occ_sched_collection,
        _lighting_sched_collection=_lighting_sched_collection,
        group_components=group_components,
        _tolerance=_tolerance,
    )
    # -- Vent. Exhaust Equip must come after zones are instantiated, since these
    # -- devices are down at the zone level instead of up at the Variant level.
    add_exhaust_vent_devices_from_hb_rooms(new_variant, _hb_room)
    add_supportive_devices_from_hb_room(new_variant, _hb_room)
    add_renewable_devices_from_hb_room(new_variant, _hb_room)
    add_phius_certification_from_hb_room(new_variant, _hb_room)
    add_phi_certification_from_hb_room(new_variant, _hb_room)
    add_PhxPhBuildingData_from_hb_room(new_variant, _hb_room)
    add_climate_from_hb_room(new_variant, _hb_room)
    add_local_pe_conversion_factors(new_variant, _hb_room)
    add_local_co2_conversion_factors(new_variant, _hb_room)
    add_elec_equip_from_hb_room(new_variant, _hb_room)

    return new_variant



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for importing WUFI XML file data."""

from typing import Dict, Any, Union, List, Optional
import pathlib
from lxml import etree
from dataclasses import dataclass


@dataclass
class Tag:
    text: Optional[str]
    tag: str
    attrib: Optional[Dict[str, str]]


def _is_list_element(_child) -> bool:
    if "count" in getattr(_child, "attrib", ""):
        return True

    # -- WUFI NONSENSE....Why the fuck don't THESE have a 'count'?
    elif _child.tag == "PEFactorsUserDef":
        return True
    elif _child.tag == "CO2FactorsUserDef":
        return True
    return False


def xml_to_dict(element: etree._Element, _level: int = 0) -> Dict[Union[List, str], Any]:
    d = {}

    if len(element) == 0:
        # -- If its a bare element with no children, just return the text
        # -- Debug
        tag = f"{_level * '  '} {element.tag :<35}"
        attrib = getattr(element, "attrib", "")
        l = len(element)

        return {element.tag: Tag(element.text, element.tag, element.attrib)}

    for child in element:  # type: ignore
        # -- Debug
        tag = f"{_level * '  '} {child.tag :<35}"
        attrib = getattr(child, "attrib", "")
        l = len(child)
        # ---

        if len(child) == 0:
            # At the end of the a branch
            if "count" in child.attrib:
                # It is just an empty container
                d[child.tag] = []
            else:
                # It is finally an actual data item
                d[child.tag] = Tag(child.text, child.tag, child.attrib)  # .text
        elif _is_list_element(child):
            # -- Fucking WUFI... sometimes the unit data is up at the parent
            if "unit" in getattr(child, "attrib", ""):
                for _ in child:
                    _.attrib["unit"] = child.attrib.get("unit", "")

            # -- The children of this node should be in a list
            d[child.tag] = []
            for sub_child in child:
                d[child.tag].append(xml_to_dict(sub_child, _level + 1))

        else:
            d[child.tag] = xml_to_dict(child, _level + 1)
    return d


def get_WUFI_XML_file_as_dict(_file_address: pathlib.Path) -> Dict[Union[str, List], Any]:
    """Read in the WUFI-XML file and return the data as a dictionary."""

    parser = etree.XMLPullParser(recover=True, encoding="utf-8")
    with open(_file_address, mode="r", encoding="utf-8") as xml_file:
        # -- Read in chunks in case the file is large
        while True:
            chunk = xml_file.read(1024)
            if not chunk:
                break
            parser.feed(chunk)

    # Get the root element of the parsed XML
    root: etree._Element = parser.close()

    # Convert the root element and all children to a Python Dictionary
    return xml_to_dict(root)


def get_WUFI_xml_file_as_str(_file_address: pathlib.Path) -> str:
    """Read in the WUFI-XML file and return it as a string.

    Arguments:
    ----------
        _file_address (pathlib.Path): A valid file path for the WUFI-XML file to read.

    Returns:
    --------
        str: The WUFI XML text, read in from the WUFI-XML file.
    """

    with open(_file_address) as xml_file:
        data = xml_file.read()
    return data



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Window Shades."""

from __future__ import annotations
from typing import ClassVar, Union
from dataclasses import dataclass, field
import uuid


@dataclass
class PhxWindowShade:
    _count: ClassVar[int] = 0

    _identifier: Union[uuid.UUID, str] = field(init=False, default_factory=uuid.uuid4)
    id_num: int = field(init=False, default=0)
    display_name: str = "__unnamed_shade__"
    operation_mode: int = 1  # 1=Reduce overheating
    reduction_factor: float = 1.0

    # -- Extra junk
    external_emissivity: float = 0.8
    absorptivity: float = 0.0
    thermal_resistance_supplement: float = 0.0
    thermal_resistance_cavity: float = 0.0
    radiation_limit: float = 1_000.0
    exclude_weekends: bool = False

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def identifier(self) -> str:
        return str(self._identifier)

    @identifier.setter
    def identifier(self, _in: str):
        if not _in:
            return
        self._identifier = str(_in)

    def __hash__(self) -> int:
        return hash(self.identifier)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions used to cleanup / optimize Honeybee-Rooms before outputting to WUFI"""

from typing import List, Dict
from functools import reduce

try:
    from ladybug_geometry.geometry2d.pointvector import Vector2D
    from ladybug_geometry.geometry2d.polygon import Polygon2D
    from ladybug_geometry.geometry3d.plane import Plane
    from ladybug_geometry.geometry3d.face import Face3D
except ImportError as e:
    raise ImportError("\nFailed to import ladybug_geometry:\n\t{}".format(e))

try:  # import the core honeybee dependencies
    from honeybee.aperture import Aperture
    from honeybee.typing import clean_ep_string
    from honeybee import room, face
    from honeybee.properties import FaceProperties
    from honeybee_energy import shw
    from honeybee.boundarycondition import Outdoors, Ground, Surface
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))

try:
    from honeybee_energy.load import people, equipment, infiltration
    from honeybee_energy.load.infiltration import Infiltration
    from honeybee_energy.properties.room import RoomEnergyProperties
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_energy:\n\t{}".format(e))

try:
    from honeybee_ph.properties.room import RoomPhProperties
    from honeybee_ph.foundations import PhFoundation
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph:\n\t{}".format(e))

try:
    from honeybee_energy_ph.properties.load.equipment import ElectricEquipmentPhProperties
    from honeybee_energy_ph.properties.hot_water.hw_system import SHWSystemPhProperties
    from honeybee_energy_ph.properties.load.people import PeoplePhProperties, PhDwellings
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_energy_ph:\n\t{}".format(e))

try:
    from PHX.model import project
    from PHX.from_HBJSON.cleanup_merge_faces import sort_hb_faces, merge_hb_faces
except ImportError as e:
    raise ImportError("\nFailed to import PHX:\n\t{}".format(e))


def _dup_face(_hb_face: face.Face) -> face.Face:
    """Duplicate a Honeybee Face and all it's properties.

    Arguments:
    ----------
        *_hb_face: (face.Face) The face to copy

    Returns:
    --------
        * (face.Face): The duplicate face.
    """

    new_face = _hb_face.duplicate()  # type: face.Face # type: ignore
    new_face._properties._duplicate_extension_attr(_hb_face._properties)

    # -- Note, this is required if the user has set custom .energy constructions
    # -- or other custom face-specific attributes
    # -- Duplicate any extensions like .ph, .energy or .radiance
    face_prop: FaceProperties = _hb_face.properties  # type: ignore
    for extension_name in face_prop._extension_attributes:
        # -- use try except to avoid the bug introduced in HB v1.5
        # TODO: fix Honeybee and then remove this try...except...
        try:
            original_extension = getattr(_hb_face._properties, f"_{extension_name}", None)
            new_extension = original_extension.duplicate()  # type: ignore
            setattr(new_face._properties, f"_{extension_name}", new_extension)
        except:
            pass

    return new_face


def _get_thermal_envelope_faces(
    _hb_room: room.Room, _all_room_ids: List[str]
) -> List[face.Face]:
    """Return a list of all the thermal-envelope faces for a Honeybee Room.

    This will include all the 'Outdoor', 'Ground' and 'Adiabatic' faces.
    If an 'Surface' faces are found, they will only be included IF they do NOT
    have exposure to another HB-Room in the set. This will omit any 'interior'
    surfaces, but will retain any surfaces which are exposed to an HB-Room which
    is part of another set (building segment). ie: the floor of a residential
    tower which is exposed to the ceiling of a commercial podium.

    Arguments:
    ----------
        *_hb_room: (room.Room) The room to get the faces from.
        * _all_room_ids: List[str] The list of all the room-group (building segment) IDs

    Returns:
    --------
        * (List[face.Face]) A List of all the thermal boundary faces.
    """

    exposed_faces: List[face.Face] = []
    for original_face in _hb_room.faces:
        # -- If it is a Surface exposure, but the face's adjacent zone is NOT part of
        # -- the room group, it must be exposed to another zone (ie: the floor surface
        # -- of a residential tower exposed to the commercial zone below). So
        face_bc = original_face.boundary_condition
        if isinstance(face_bc, Surface):
            adjacent_room_name = face_bc.boundary_condition_objects[-1]
            if adjacent_room_name in _all_room_ids:
                continue

        exposed_faces.append(_dup_face(original_face))

    return exposed_faces


def _get_room_exposed_face_area(_hb_room: room.Room) -> float:
    """Returns the total surface area for all the 'exposed' faces (Ground / Outdoors) of a HB-Room
    For Phius, infiltration-exposed surfaces include all, including 'Ground'
    unlike for Honeybee where only 'Outdoors' count as 'exposed'

    Arguments:
    ----------
        *_hb_room: (room.Room) The Honeybee Room to get the surfaces of.

    Returns:
    --------
        * (float) The total surface area.
    """

    return sum(
        fc.area
        for fc in _hb_room.faces
        if isinstance(fc.boundary_condition, (Outdoors, Ground))
    )


def all_unique_ph_dwelling_objects(_hb_rooms: List[room.Room]) -> List[PhDwellings]:
    """Return a list of all the unique PhDwelling objects from a set of HB-Rooms.

    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): A list of the HB-Rooms

    Returns:
    --------
        * (List[PhDwellings])
    """
    dwellings = {
        room.properties.energy.people.properties.ph.dwellings  # type: ignore
        for room in _hb_rooms
        if room.properties.energy.people  # type: ignore
    }
    return list(dwellings)


def merge_occupancies(_hb_rooms: List[room.Room]) -> people.People:
    """Returns a new HB-People-Obj with it's values set from a list of input HB-Rooms.

    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): A list of the HB-Rooms to build the merged
            HB-People object from.

    Return:
    -------
        * (people.People): A new Honeybee People object with values merged from the HB-Rooms.
    """

    # -------------------------------------------------------------------------
    # Calculate the total dwelling-unit count. Ensure always at least 1 (even Non-Res)
    total_ph_dwellings = 0
    for dwelling_obj in all_unique_ph_dwelling_objects(_hb_rooms):
        total_ph_dwellings += int(dwelling_obj.num_dwellings)
    merged_ph_dwellings = PhDwellings(_num_dwellings=max(total_ph_dwellings, 1))

    # -------------------------------------------------------------------------
    # -- Merge all the Rooms
    total_ph_bedrooms = 0
    total_ph_people = 0.0
    total_hb_people = 0.0
    for hb_room in _hb_rooms:
        # -- Type Aliases
        hb_room_prop_energy: RoomEnergyProperties = hb_room.properties.energy  # type: ignore
        hb_ppl_obj = hb_room_prop_energy.people

        # -- Sometimes there is no 'People'
        if hb_ppl_obj is None:
            continue

        hbph_people_prop_ph: PeoplePhProperties = hb_ppl_obj.properties.ph  # type: ignore
        total_ph_bedrooms += int(hbph_people_prop_ph.number_bedrooms)
        total_ph_people += float(hbph_people_prop_ph.number_people)
        total_hb_people += hb_ppl_obj.people_per_area * hb_room.floor_area

    # -------------------------------------------------------------------------
    # -- Build up the new People object's attributes
    total_floor_area = sum(rm.floor_area for rm in _hb_rooms)
    new_hb_prop_energy = _hb_rooms[0].properties.energy  # type: RoomEnergyProperties
    new_hb_ppl = new_hb_prop_energy.people.duplicate()  # type: people.People
    new_hb_ppl_prop_ph = new_hb_ppl.properties.ph  # type: PeoplePhProperties
    new_hb_ppl.people_per_area = total_hb_people / total_floor_area
    new_hb_ppl_prop_ph.number_bedrooms = total_ph_bedrooms
    new_hb_ppl_prop_ph.number_people = total_ph_people
    new_hb_ppl_prop_ph.dwellings = merged_ph_dwellings

    return new_hb_ppl


def merge_infiltrations(_hb_rooms: List[room.Room]) -> infiltration.Infiltration:
    """Returns a new HB-Infiltration-Obj with it's values set from a list of input HB-Rooms.

    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): A list of the HB-Rooms to build the merged
            HB-Infiltration object from.

    Return:
    -------
        * (infiltration.Infiltration): A new Honeybee Infiltration object
            with values merged from the HB-Rooms.
    """

    # -- Calculate the total airflow per room, total exposed area per room
    total_m3_s = 0.0
    total_exposed_area = 0.0
    for room in _hb_rooms:
        room_infil_exposed_area = _get_room_exposed_face_area(room)
        room_prop_energy: RoomEnergyProperties = room.properties.energy  # type: ignore
        room_infil_m3_s = (
            room_infil_exposed_area * room_prop_energy.infiltration.flow_per_exterior_area
        )

        total_exposed_area += room_infil_exposed_area
        total_m3_s += room_infil_m3_s

    # -- Set the new Infiltration Object's attr to the weighted average
    reference_room = _hb_rooms[0]
    reference_room_prop_energy: RoomEnergyProperties = reference_room.properties.energy  # type: ignore
    new_infil: Infiltration = reference_room_prop_energy.infiltration.duplicate()  # type: ignore
    try:
        new_infil.flow_per_exterior_area = total_m3_s / total_exposed_area
    except ZeroDivisionError:
        new_infil.flow_per_exterior_area = 0.0

    return new_infil


def merge_shw(_hb_rooms: List[room.Room]) -> shw.SHWSystem:
    """Merge together several HB-Room's Honeybee-Energy SHW System objects (if they exist).

    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): The list of Honeybee Rooms to get the
            SHW System System from.

    Returns:
    --------
        * (shw.SHWSystem): A single new Honeybee-Energy SHW System Object.
    """

    # -- Find the first HBE SHWSystem System in the list of HB-Rooms, and use that as the 'base'
    for room in _hb_rooms:
        room_prop_energy: RoomEnergyProperties = room.properties.energy  # type: ignore
        if room_prop_energy.shw:
            new_shw: shw.SHWSystem = room_prop_energy.shw.duplicate()
            break
    else:
        # -- If no SHWSystem found anywhere, return a new default HBE SHWSystem System
        return shw.SHWSystem(
            identifier=clean_ep_string("_default_shw_system_"),
            equipment_type="Electric_WaterHeater",
        )

    # -- Merge the SHWSystemPhProperties and then apply it to the new HBE SHWSystem System
    shw_props: List[SHWSystemPhProperties] = [
        hb_room.properties.energy.shw.properties.ph  # type: ignore
        for hb_room in _hb_rooms
        if hb_room.properties.energy.shw  # type: ignore
    ]

    new_prop = reduce(lambda a, b: a + b, shw_props)
    new_shw.properties._ph = new_prop  # type: ignore

    return new_shw


def merge_elec_equip(_hb_rooms: List[room.Room]) -> equipment.ElectricEquipment:
    """Returns a new HB-ElectricEquipment-Obj with it's values set from a list of input HB-Rooms.

    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): A list of the HB-Rooms to build the merged
            HB-ElectricEquipment object from.

    Return:
    -------
        * (equipment.ElectricEquipment): A new Honeybee ElectricEquipment object
            with values merged from the HB-Rooms.
    """

    # -- Collect all the unique PH-Equipment in all the rooms.
    # -- Increase the quantity for each duplicate piece of equipment found.
    ph_equipment = {}
    for room in _hb_rooms:
        room_prop_energy: RoomEnergyProperties = room.properties.energy  # type: ignore
        room_ee_prop = room_prop_energy.electric_equipment.properties
        room_ee_prop_ph: ElectricEquipmentPhProperties = room_ee_prop.ph  # type: ignore

        for equip_key, equip in room_ee_prop_ph.equipment_collection.items():
            try:
                ph_equipment[equip_key].quantity += 1
            except KeyError:
                ph_equipment[equip_key] = equip

    # -- Calculate the total Watts of elec-equipment, total floor-area
    total_floor_area = sum(rm.floor_area for rm in _hb_rooms)
    total_watts = sum(
        (rm.floor_area * rm.properties.energy.electric_equipment.watts_per_area)  # type: ignore
        for rm in _hb_rooms
    )

    # -- Build a new HB-Elec-Equip from the reference room, add all the PH-Equipment to it.
    reference_room = _hb_rooms[0]
    reference_room_prop_energy: RoomEnergyProperties = reference_room.properties.energy  # type: ignore
    ref_room_ee: equipment.ElectricEquipment = (
        reference_room_prop_energy.electric_equipment
    )

    new_hb_equip: equipment.ElectricEquipment = ref_room_ee.duplicate()  # type: ignore
    new_hb_equip.watts_per_area = total_watts / total_floor_area
    new_hb_equip_prop_ph: ElectricEquipmentPhProperties = new_hb_equip.properties.ph  # type: ignore
    new_hb_equip_prop_ph.equipment_collection.remove_all_equipment()

    for ph_item in ph_equipment.values():
        new_hb_equip_prop_ph.equipment_collection.add_equipment(ph_item)

    return new_hb_equip


def check_room_has_spaces(_hb_room: room.Room) -> None:
    rm_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    if len(rm_prop_ph.spaces) == 0:
        print(
            f"Warning: Room '{_hb_room.display_name}' has no spaces?"
            " Merge may not work correctly"
        )


def merge_foundations(_hb_rooms: List[room.Room]) -> Dict[str, PhFoundation]:
    """

    Arguments:
    ----------
        *

    Returns:
    --------
        *
    """

    # -- Group by Identifier
    foundation_groups = {}
    for rm in _hb_rooms:
        for foundation in rm.properties.ph.ph_foundations:
            foundation_groups[foundation.identifier] = foundation

    # -- Warn if more than 3 of them
    if len(foundation_groups) > 3:
        msg = (
            f"\tWarning: WUFI-Passive only allows 3 Foundation types. "
            f" {len(foundation_groups)} found on the Building Segment '"
            f"{_hb_rooms[0].properties.ph.ph_bldg_segment.display_name}'?"
        )
        print(msg)

    return foundation_groups


def merge_rooms(
    _hb_rooms: List[room.Room],
    _tolerance: float,
    _angle_tolerance_degrees: float,
    _merge_faces: bool = False,
) -> room.Room:
    """Merge together a group of Honeybee Rooms into a new single HB Room.

    This will ignore any 'interior' Honeybee-Faces with a 'Surface' boundary
    condition and will only keep the 'exposed' Honeybee Faces with boundary_conditions
    of 'Outdoors', 'Ground' and 'Adiabatic' to build the new Honeybee Room from.

    Arguments:
    ----------
        * _hb_rooms (list[room.Room]):

    Returns:
    --------
        * room.Room: The new Honeybee Room.
    """
    reference_room = _hb_rooms[0]

    # -------------------------------------------------------------------------
    # Collect all the HB-Room names for Surface adjacency checking
    all_hb_room_identifiers = [rm.identifier for rm in _hb_rooms]

    # -------------------------------------------------------------------------
    # -- Get only the 'exposed' thermal envelope faces to build a new HB-Room with
    exposed_faces = []
    for hb_room in _hb_rooms:
        exposed_faces += _get_thermal_envelope_faces(hb_room, all_hb_room_identifiers)

    # -------------------------------------------------------------------------
    # -- Try and merge the Faces to simplify the geometry
    if _merge_faces:
        face_groups = sort_hb_faces(exposed_faces, _tolerance, _angle_tolerance_degrees)
        merged_faces = []
        for face_group in face_groups:
            merged_faces.extend(
                merge_hb_faces(face_group, _tolerance, _angle_tolerance_degrees)
            )
        exposed_faces = merged_faces

    # -------------------------------------------------------------------------
    # Build the new room from the exposed (possibly merged) faces
    new_room = room.Room(
        identifier=reference_room.properties.ph.ph_bldg_segment.display_name,  # type: ignore
        faces=exposed_faces,
    )

    # -------------------------------------------------------------------------
    # -- Set the new Merged-Room's properties.ph and
    # -- properties.energy to match the 'reference' room to start with, but leave
    # -- off the PH-Spaces so they don't get duplicated
    ref_rm_prop_ph: RoomPhProperties = reference_room.properties.ph  # type: ignore
    new_rm_prop_ph: RoomPhProperties = new_room.properties.ph  # type: ignore
    dup_ph_prop = ref_rm_prop_ph.duplicate(new_rm_prop_ph, include_spaces=False)
    dup_ph_prop._ph_foundations = merge_foundations(_hb_rooms)
    setattr(new_room._properties, "_ph", dup_ph_prop)

    ref_rm_prop_energy: RoomEnergyProperties = reference_room.properties.energy  # type: ignore
    new_rm_prop_energy: RoomEnergyProperties = new_room.properties.energy  # type: ignore
    dup_energy_prop = ref_rm_prop_energy.duplicate(new_rm_prop_energy)
    setattr(new_room._properties, "_energy", dup_energy_prop)

    # -------------------------------------------------------------------------
    # -- Then, collect all the spaces from the input rooms and add to the NEW room
    # -- NOTE: this has to be done AFTER the duplicate()
    # -- call, otherwise not all the spaces will transfer over properly.
    for hb_room in _hb_rooms:
        check_room_has_spaces(hb_room)

        rm_prop_ph: RoomPhProperties = hb_room.properties.ph  # type: ignore
        for existing_space in rm_prop_ph.spaces:
            # -- Preserve the space host HB-Room's .energy and .ph properties over
            # -- on the space itself. We need to do this cus' the host HB-Room is being
            # -- removed and we want to preserve HVAC and program info for the spaces.
            existing_space.properties._energy = hb_room.properties._energy.duplicate(  # type: ignore
                new_host=existing_space
            )
            # TODO: Verify that this can be removed without causing any issues?
            # Note: it it also wrong - RoomPhProperties are being applied over SpacePhProperties.
            # existing_space.properties._ph = hb_room.properties._ph.duplicate(
            #     new_host=existing_space)

            new_rm_prop_ph: RoomPhProperties = new_room.properties.ph  # type: ignore
            new_rm_prop_ph.add_new_space(existing_space)

    # -------------------------------------------------------------------------
    # -- Merge the hb_rooms' load values
    new_rm_prop_energy: RoomEnergyProperties = new_room.properties.energy  # type: ignore
    new_rm_prop_energy.infiltration = merge_infiltrations(_hb_rooms)
    new_rm_prop_energy.people = merge_occupancies(_hb_rooms)
    new_rm_prop_energy.electric_equipment = merge_elec_equip(_hb_rooms)
    new_rm_prop_energy.shw = merge_shw(_hb_rooms)

    # -------------------------------------------------------------------------
    # -- TODO: Can I merge together the surfaces as well?
    # -- For larger models, I think this will be important.... hmm....
    # -- Organize the surfaces by -> assembly / exposure / orientation (normal)

    return new_room


def weld_vertices(_variant: project.PhxVariant) -> project.PhxVariant:
    """
    Used to try and weld/unify the vertices of a variant.

    This is helpful to reduce the complexity / number of variants in a large
    model so that WUFI can actually open it.

    Arguments:
    ----------
        * _variant (project.Variant): The Variant object to weld the vertices for.

    Returns:
    --------
        * (project.Variant): The variant, with its vertix objects welded.

    """

    unique_vertix_dict = {}
    for component in _variant.building.all_components:
        for polygon in component.polygons:
            for i, vert in enumerate(polygon.vertices):
                try:
                    # -- See if the vertex is already in the dict,
                    # --if so, use that one.
                    vertex_from_dict = unique_vertix_dict[vert.unique_key]
                    polygon.set_vertex(vertex_from_dict, i)
                except KeyError:
                    # -- If the vertex is not in the dict, add it.
                    unique_vertix_dict[vert.unique_key] = vert

    return _variant



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "IHG non-res" worksheet."""

from __future__ import annotations
from typing import List, Optional

from PHX.xl import xl_data
from PHX.xl.xl_data import col_offset
from PHX.PHPP.phpp_model import uvalues_constructor
from PHX.PHPP.phpp_localization import shape_model
from PHX.xl import xl_app


class IhgNonRes:
    """IO Controller for the PHPP "IHG non-res" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.IhgNonRes):
        self.xl = _xl
        self.shape = _shape



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Data-entry constructor for the U-Values Worksheet."""

from dataclasses import dataclass
from typing import List
from functools import partial

from PHX.model.constructions import PhxConstructionOpaque, PhxLayer

from PHX.xl import xl_data
from PHX.PHPP.phpp_localization import shape_model


@dataclass
class ConstructorBlock:
    """A single U-Value/Constructor entry block."""

    shape: shape_model.UValues
    phx_construction: PhxConstructionOpaque = PhxConstructionOpaque()

    def _create_range(self, _field_name: str, _row_offset: int, _start_row: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.constructor.inputs, _field_name).column
        return f"{col}{_start_row + _row_offset}"

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.constructor.inputs, _field_name).unit

    def is_mass_material(self, _layer: PhxLayer) -> bool:
        """Return True if the Layer is an SD 'Mass' Layer."""

        for material in _layer.materials:
            if material.display_name == "MAT_Mass" and material.conductivity == 100:
                return True

        return False

    def create_xl_items(self, _sheet_name: str, _start_row: int) -> List[xl_data.XlItem]:
        """Convert the PHX-Construction into a list of XLItems for writing to the PHPP."""

        create_range = partial(self._create_range, _start_row=_start_row)
        XLItemUValues = partial(xl_data.XlItem, _sheet_name)

        # -- Build the basic assembly attributes
        xl_items_list: List[xl_data.XlItem] = [
            XLItemUValues(
                create_range(
                    "display_name", self.shape.constructor.inputs.name_row_offset
                ),
                f"'{self.phx_construction.display_name}",
            ),
            XLItemUValues(
                create_range("r_si", self.shape.constructor.inputs.rse_row_offset),
                0.0,
                "M2K/W",
                self._get_target_unit("r_si"),
            ),
            XLItemUValues(
                create_range("r_se", self.shape.constructor.inputs.rsi_row_offset),
                0.0,
                "M2K/W",
                self._get_target_unit("r_se"),
            ),
        ]

        # -- Build all the layers of the assembly
        for i, layer in enumerate(
            self.phx_construction.layers,
            start=self.shape.constructor.inputs.first_layer_row_offset,
        ):
            if self.is_mass_material(layer):
                continue

            xl_items_list.append(
                XLItemUValues(
                    create_range("thickness", i),
                    layer.thickness_mm,
                    "MM",
                    self._get_target_unit("thickness"),
                )
            )

            # -- add in each of the PhxMaterials found in the PhxLayer.materials collection
            for j, material in enumerate(layer.materials, start=1):
                layer_items: List[xl_data.XlItem] = [
                    XLItemUValues(
                        create_range(f"sec_{j}_description", i),
                        f"'{material.display_name}",
                    ),
                    XLItemUValues(
                        create_range(f"sec_{j}_conductivity", i),
                        material.conductivity,
                        "W/MK",
                        self._get_target_unit(f"sec_{j}_conductivity"),
                    ),
                ]
                # -- Only add the percentage info for sections 2 and 3...
                if j > 1:
                    layer_items.append(
                        XLItemUValues(
                            create_range(
                                f"sec_{j}_percentage",
                                getattr(
                                    self.shape.constructor.inputs, f"sec_{j}_percentage"
                                ).row,
                            ),
                            material.percentage_of_assembly,
                        )
                    )
                xl_items_list.extend(layer_items)

        return xl_items_list



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP 'Components' worksheet."""

from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional, Set, Dict

from ph_units.unit_type import Unit

from PHX.xl import xl_app
from PHX.xl.xl_data import col_offset
from PHX.PHPP.phpp_model.component_vent import VentilatorRow
from PHX.PHPP.phpp_model.component_glazing import GlazingRow
from PHX.PHPP.phpp_model.component_frame import FrameRow
from PHX.PHPP.phpp_localization import shape_model


@dataclass
class ExistingGlazingTypeData:
    name: str
    g_value: Unit
    u_value: Unit

    @property
    def key(self) -> str:
        return f"{self.name}-{self.g_value}-{self.u_value}"


class Glazings:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Components):
        self.xl = _xl
        self.shape = _shape
        self._section_header_row: Optional[int] = None
        self._section_first_entry_row: Optional[int] = None
        self._section_last_entry_row: Optional[int] = None
        self.cache = {}

    @property
    def section_header_row(self) -> int:
        """Return the row number of the Glazings section header."""
        if not self._section_header_row:
            self._section_header_row = self.find_section_header_row()
        return self._section_header_row

    @property
    def section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the Glazing input section."""
        if not self._section_first_entry_row:
            self._section_first_entry_row = self.find_section_first_entry_row()
        return self._section_first_entry_row

    @property
    def section_last_entry_row(self) -> int:
        """Return the row number of the very last user-input entry row in the Glazing input section."""
        if not self._section_last_entry_row:
            self._section_last_entry_row = self.find_section_last_entry_row()
        return self._section_last_entry_row

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        """Return the row number of the Glazings section header."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.glazings.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data):
            if self.shape.glazings.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.glazings.locator_string_header}'"
            f"header on the '{self.shape.name}' sheet, column {self.shape.glazings.locator_col_header}?"
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the Glazing input section."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.glazings.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row + 25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            if val == self.shape.glazings.locator_string_entry:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.glazings.locator_string_entry}' "
            f" entry start on the '{self.shape.name}' sheet, column {self.shape.glazings.locator_col_entry}?"
        )

    def find_section_last_entry_row(self, _start_row: Optional[int] = None) -> int:
        """Return the last row of the glazing input section."""
        if not _start_row:
            _start_row = self.section_first_entry_row
        elif _start_row > 10_000:
            raise Exception(
                f"Error: Cannot find the last row in the '{self.shape.name}' sheet, column {self.shape.glazings.locator_col_entry}?"
            )

        _row_end = _start_row + 500
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.glazings.locator_col_entry,
            _row_start=_start_row,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=_start_row):
            if not val:
                return i - 1
        else:
            return self.find_section_last_entry_row(_row_end)

    def find_first_empty_row(self) -> int:
        """Return the first empty row in the glazing input section."""
        search_col = str(self.shape.glazings.inputs.description.column)
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=search_col,
            _row_start=self.section_first_entry_row,
            _row_end=self.section_last_entry_row,
        )

        for i, val in enumerate(xl_data, start=self.section_first_entry_row):
            if not val:
                return i

        raise Exception(
            f"Error: Cannot find the first empty row in the '{self.shape.name}' sheet, column {search_col}?"
        )

    def get_glazing_phpp_id_by_name(
        self, _name: str, _use_cache: bool = False
    ) -> Optional[str]:
        """Return the PHPP Glazing ID for the given name."""
        if _use_cache:
            try:
                return self.cache[_name]
            except KeyError:
                pass

        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=1,
            row_end=500,
            col=str(self.shape.glazings.inputs.description.column),
            find=_name,
        )

        if not row:
            return

        prefix = self.xl.get_data(
            self.shape.name,
            f"{col_offset(str(self.shape.glazings.inputs.description.column), -1)}{row}",
        )
        print(f"Getting PHPP Glazing PHPP-id for {_name}")
        name_with_id = f"{prefix}-{_name}"

        self.cache[_name] = name_with_id

        return name_with_id

    def get_glazing_phpp_id_by_row_num(self, _row_num: int) -> str:
        """Return the PHPP Glazing ID ("01ud-MyGlass", etc..) for the given row number."""
        id_col = str(self.shape.glazings.inputs.id.column)
        id_num = self.xl.get_data(self.shape.name, f"{id_col}{_row_num}")
        name_col = str(self.shape.glazings.inputs.description.column)
        id_name = self.xl.get_data(self.shape.name, f"{name_col}{_row_num}")
        return f"{id_num}-{id_name}"

    def get_all_glazing_types(self) -> List[ExistingGlazingTypeData]:
        """Return a set of all glazing types in the Glazing input section."""
        glazing_types: Dict[str, ExistingGlazingTypeData] = {}
        unit_type = str(self.shape.glazings.inputs.u_value.unit)
        start = f"{self.shape.glazings.inputs.description.column}{self.section_first_entry_row}"
        end = f"{self.shape.glazings.inputs.u_value.column}{self.section_last_entry_row}"

        data = self.xl.get_data(self.shape.name, f"{start}:{end}")

        if not data:
            return []

        for row in data:
            if "None" in row or None in row:
                continue

            exiting_glazing_type = ExistingGlazingTypeData(
                str(row[0]),
                Unit(row[1], "-"),
                Unit(row[2], unit_type),
            )
            glazing_types[exiting_glazing_type.key] = exiting_glazing_type

        return [glazing_types[k] for k in sorted(glazing_types.keys())]


class Frames:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Components):
        self.xl = _xl
        self.shape = _shape
        self._section_header_row: Optional[int] = None
        self._section_first_entry_row: Optional[int] = None
        self._section_last_entry_row: Optional[int] = None
        self.cache = {}

    @property
    def section_header_row(self) -> int:
        """Return the row number of the 'Frames' section header."""
        if not self._section_header_row:
            self._section_header_row = self.find_section_header_row()
        return self._section_header_row

    @property
    def section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the Frames input section."""
        if not self._section_first_entry_row:
            self._section_first_entry_row = self.find_section_first_entry_row()
        return self._section_first_entry_row

    @property
    def section_last_entry_row(self) -> int:
        """Return the row number of the very last user-input entry row in the Frames input section."""
        if not self._section_last_entry_row:
            self._section_last_entry_row = self.find_section_last_entry_row()
        return self._section_last_entry_row

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.frames.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end,
        )
        """Return the row number of the 'Frames' section header."""

        for i, val in enumerate(xl_data):
            if self.shape.frames.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.frames.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.frames.locator_col_header}?"
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the Frames input section."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.frames.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row + 25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            if val == self.shape.frames.locator_string_entry:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.frames.locator_string_entry}'"
            f"entry start on the 'Components' sheet, column {self.shape.frames.locator_col_entry}?"
        )

    def find_section_last_entry_row(self, _start_row: Optional[int] = None) -> int:
        """Return the last row of the Frames input section."""
        if not _start_row:
            _start_row = self.section_first_entry_row
        elif _start_row > 10_000:
            raise Exception(
                f"Error: Cannot find the last row in the '{self.shape.name}' sheet, column {self.shape.frames.locator_col_entry}?"
            )

        _row_end = _start_row + 500
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.frames.locator_col_entry,
            _row_start=_start_row,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=self.section_first_entry_row):
            if not val:
                return i - 1
        else:
            return self.find_section_last_entry_row(_row_end)

    def find_first_empty_row(self) -> int:
        """Return the first empty row in the frames input section."""
        search_col = str(self.shape.frames.inputs.description.column)
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=search_col,
            _row_start=self.section_first_entry_row,
            _row_end=self.section_last_entry_row,
        )

        for i, val in enumerate(xl_data, start=self.section_first_entry_row):
            if not val:
                return i

        raise Exception(
            f"Error: Cannot find the first empty row in the '{self.shape.name}' sheet, column {search_col}?"
        )

    def get_frame_phpp_id_by_name(
        self,
        _name: str,
        _row_start: int = 1,
        _row_end: int = 500,
        _use_cache: bool = False,
    ) -> str:
        """Return the PHPP ID of a Frame component by name."""
        # -- Try and use the Cached value first
        if _use_cache:
            try:
                return self.cache[_name]
            except KeyError:
                pass

        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=_row_start,
            row_end=_row_end,
            col=str(self.shape.frames.inputs.description.column),
            find=_name,
        )

        if not row:
            msg = (
                f'Error: Cannot find a Frame component named: "{_name}" in'
                f"column {self.shape.frames.inputs.description.column}?"
            )
            raise Exception(msg)
        prefix = self.xl.get_data(
            self.shape.name,
            f"{col_offset(str(self.shape.frames.inputs.description.column), -1)}{row}",
        )
        print(f"Getting PHPP Frame id for {_name}")
        name_with_id = f"{prefix}-{_name}"
        self.cache[_name] = name_with_id

        return name_with_id

    def get_frame_phpp_id_by_row_num(self, _row_num: int) -> str:
        """Return the PHPP Frame ID ("01ud-MyFrame", etc..) for the given row number."""
        id_col = str(self.shape.frames.inputs.id.column)
        id_num = self.xl.get_data(self.shape.name, f"{id_col}{_row_num}")
        name_col = str(self.shape.frames.inputs.description.column)
        id_name = self.xl.get_data(self.shape.name, f"{name_col}{_row_num}")
        return f"{id_num}-{id_name}"


class Ventilators:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Components):
        self.xl = _xl
        self.shape = _shape
        self._section_header_row: Optional[int] = None
        self._section_first_entry_row: Optional[int] = None
        self._section_last_entry_row: Optional[int] = None

    @property
    def section_header_row(self) -> int:
        """Return the row number of the 'Ventilators' section header."""
        if not self._section_header_row:
            self._section_header_row = self.find_section_header_row()
        return self._section_header_row

    @property
    def section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the Ventilators input section."""
        if not self._section_first_entry_row:
            self._section_first_entry_row = self.find_section_first_entry_row()
        return self._section_first_entry_row

    @property
    def section_last_entry_row(self) -> int:
        """Return the row number of the very last user-input entry row in the Ventilators input section."""
        if not self._section_last_entry_row:
            self._section_last_entry_row = self.find_section_last_entry_row()
        return self._section_last_entry_row

    def find_section_last_entry_row(self, _start_row: Optional[int] = None) -> int:
        """Return the last row of the Ventilators input section."""
        if not _start_row:
            _start_row = self.section_first_entry_row
        elif _start_row > 10_000:
            raise Exception(
                f"Error: Cannot find the last row in the '{self.shape.name}' sheet, column {self.shape.ventilators.locator_col_entry}?"
            )

        _row_end = _start_row + 500
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.ventilators.locator_col_entry,
            _row_start=_start_row,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=self.section_first_entry_row):
            if not val:
                return i - 1
        else:
            return self.find_section_last_entry_row(_row_end)

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        """Return the row number of the 'Ventilators' section header."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.ventilators.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data):
            if self.shape.ventilators.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.ventilators.locator_string_header}' header on the "
            f"'{self.shape.name}' sheet, column {self.shape.ventilators.locator_col_header}?"
        )

    def find_section_first_entry_row(self) -> int:
        """Return the first row of the Ventilators input section."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.ventilators.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row + 25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            if val == self.shape.ventilators.locator_string_entry:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.ventilators.locator_string_entry}' entry start on "
            f"the '{self.shape.name}' sheet, column {self.shape.ventilators.locator_col_entry}?"
        )

    def find_first_empty_row(self) -> int:
        """Return the first empty row in the Ventilators input section."""
        search_col = str(self.shape.ventilators.inputs.display_name.column)
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=search_col,
            _row_start=self.section_first_entry_row,
            _row_end=self.section_last_entry_row,
        )

        for i, val in enumerate(xl_data, start=self.section_first_entry_row):
            if not val:
                return i

        raise Exception(
            f"Error: Cannot find the first empty row in the '{self.shape.name}' sheet, column {search_col}?"
        )

    def get_ventilator_phpp_id_by_name(
        self, _name: str, _row_start: int = 1, _row_end: int = 500
    ) -> str:
        """Return the PHPP ID of a Ventilator component by name."""
        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=_row_start,
            row_end=_row_end,
            col=str(self.shape.ventilators.inputs.display_name.column),
            find=_name,
        )

        if not row:
            raise Exception(
                f'Error: Cannot find a Ventilator component named: "{_name}"]'
                f"in column {self.shape.ventilators.inputs.display_name.column}?"
            )

        prefix = self.xl.get_data(
            self.shape.name,
            f"{col_offset(str(self.shape.ventilators.inputs.display_name.column), -1)}{row}",
        )
        return f"{prefix}-{_name}"

    def get_ventilator_phpp_id_by_row_num(self, _row_num: int) -> str:
        """Return the PHPP Ventilator ID ("01ud-MyVentilator", etc..) for the given row number."""
        id_col = str(self.shape.ventilators.inputs.id.column)
        id_num = self.xl.get_data(self.shape.name, f"{id_col}{_row_num}")
        name_col = str(self.shape.ventilators.inputs.display_name.column)
        id_name = self.xl.get_data(self.shape.name, f"{name_col}{_row_num}")
        return f"{id_num}-{id_name}"


class Components:
    """IO Controller for PHPP "Components" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, shape: shape_model.Components):
        self.xl = _xl
        self.shape = shape
        self.glazings = Glazings(self.xl, self.shape)
        self.frames = Frames(self.xl, self.shape)
        self.ventilators = Ventilators(self.xl, self.shape)

    @property
    def first_empty_glazing_row_num(self) -> int:
        """Return the row number of the first empty row in the Glazings section."""
        return self.glazings.find_first_empty_row()

    @property
    def first_empty_frame_row_num(self) -> int:
        """Return the row number of the first empty row in the Frames section."""
        return self.frames.find_first_empty_row()

    def write_single_glazing(self, _row_num: int, _glazing_row: GlazingRow) -> str:
        """Write a single GlazingRow object to the PHPP "Components" worksheet.

        Return:
        ------
            * (str): The PHPP ID-name of the glazing component written to the PHPP.
        """

        for item in _glazing_row.create_xl_items(self.shape.name, _row_num=_row_num):
            self.xl.write_xl_item(item)
        return self.glazings.get_glazing_phpp_id_by_row_num(_row_num)

    def write_glazings(self, _glazing_rows: List[GlazingRow]) -> None:
        """Write a list of GlazingRow objects to the PHPP "Components" worksheet."""
        for i, glazing_row in enumerate(
            _glazing_rows, start=self.glazings.section_first_entry_row
        ):
            self.write_single_glazing(i, glazing_row)

    def write_single_frame(self, _row_num: int, _frame_row: FrameRow) -> str:
        """Write a single FrameRow object to the PHPP "Components" worksheet.

        Return:
        ------
            * (str): The PHPP ID-name of the frame component written to the PHPP.
        """
        for item in _frame_row.create_xl_items(self.shape.name, _row_num=_row_num):
            self.xl.write_xl_item(item)
        return self.frames.get_frame_phpp_id_by_row_num(_row_num)

    def write_frames(self, _frame_row: List[FrameRow]) -> None:
        """Write a list of FrameRow objects to the PHPP "Components" worksheet."""
        start = self.frames.section_first_entry_row
        for i, frame_row in enumerate(_frame_row, start=start):
            self.write_single_frame(i, frame_row)

    def write_single_ventilator(
        self, _row_num: int, _ventilator_row: VentilatorRow
    ) -> str:
        """Write a single VentilatorRow object to the PHPP "Components" worksheet.

        Return:
        ------
            * (str): The PHPP ID-name of the ventilator component written to the PHPP.
        """
        for item in _ventilator_row.create_xl_items(self.shape.name, _row_num=_row_num):
            self.xl.write_xl_item(item)
        return self.ventilators.get_ventilator_phpp_id_by_row_num(_row_num)

    def write_ventilators(self, _ventilator_row: List[VentilatorRow]) -> None:
        """Write a list of VentilatorRow objects to the PHPP "Components" worksheet."""
        start = self.ventilators.section_first_entry_row
        for i, ventilator_row in enumerate(_ventilator_row, start):
            self.write_single_ventilator(i, ventilator_row)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to build PHX-Foundations from HBPH source objects."""

from typing import Dict
from honeybee_ph.foundations import PhFoundation

from PHX.model.enums.foundations import FoundationType
from PHX.model import ground

def create_phx_foundation_from_hbph(_hbph_foundation: PhFoundation) -> ground.PhxFoundation:
    """Return a new PhxFoundation object with attributes based on a source HBPH-Foundation.
    
    Arguments:
    ----------
        *
    
    Returns:
    --------
        * (PhxFoundation)
    """

    phx_foundation_type_map: Dict[str, ground.PhxFoundationTypes] = {
        "PhHeatedBasement": ground.PhxHeatedBasement,
        "PhUnheatedBasement": ground.PhxUnHeatedBasement,
        "PhSlabOnGrade": ground.PhxSlabOnGrade,
        "PhVentedCrawlspace": ground.PhxVentedCrawlspace,
        "PhFoundation": ground.PhxFoundation,
    }
    new_phx_foundation = phx_foundation_type_map[_hbph_foundation.__class__.__name__]()
    new_phx_foundation.display_name = _hbph_foundation.display_name
    new_phx_foundation.foundation_type_num = FoundationType(_hbph_foundation.foundation_type.number)


    # -- Pull out all the PH attributes and set the PHX ones to match.
    for attr_name in vars(_hbph_foundation).keys():
        if str(attr_name).startswith("_"):
            attr_name = attr_name[1:]

        try:
            # try and set any Enums by number first...
            setattr(new_phx_foundation, attr_name, getattr(_hbph_foundation, attr_name).number)
        except AttributeError:
            # ... then just set copy over any non-Enum values
            try:
                setattr(new_phx_foundation, attr_name, getattr(_hbph_foundation, attr_name))
            except KeyError:
                raise 
            except Exception as e:
                msg = f"\n\tError setting attribute '{attr_name}' on '{new_phx_foundation.__class__.__name__}'?"
                raise Exception(msg)

    return new_phx_foundation



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a PHPP Windows/Window-Entry row"""

from dataclasses import dataclass
from typing import List, Tuple, Optional
from functools import partial

from PHX.model import constructions, geometry

from PHX.xl import xl_data
from PHX.PHPP.phpp_localization import shape_model


def get_name_from_glazing_id(
    _phpp_glazing_id: Optional[xl_data.xl_range_single_value],
) -> str:
    """Return the aperture's PHPP-Name (ie: "MyGlass") from phpp-id-string."""
    try:
        return str(_phpp_glazing_id).split("-", 1)[1]
    except:
        if _phpp_glazing_id == None or _phpp_glazing_id == "None":
            return ""
        else:
            msg = (
                f"Error getting construction PHPP-Name? "
                f"Could not split {_phpp_glazing_id} on '-'?"
            )
            raise Exception(msg)


@dataclass
class WindowRow:
    """Model class for a single Window entry row."""

    __slots__ = (
        "shape",
        "phx_polygon",
        "phx_construction",
        "phpp_host_surface_id_name",
        "phpp_id_frame",
        "phpp_id_glazing",
        "phpp_id_variant_type",
    )
    shape: shape_model.Windows
    phx_polygon: geometry.PhxPolygonRectangular
    phx_construction: constructions.PhxConstructionWindow
    phpp_host_surface_id_name: Optional[str]
    phpp_id_frame: Optional[str]
    phpp_id_glazing: Optional[str]
    phpp_id_variant_type: Optional[str]

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.window_rows.inputs, _field_name).column
        return f"{col}{_row_num}"

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.window_rows.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemWindows = partial(xl_data.XlItem, _sheet_name)
        items: List[xl_data.XlItem] = [
            XLItemWindows(create_range("variant_input"), self.phpp_id_variant_type),
            XLItemWindows(create_range("quantity"), 1),
            XLItemWindows(
                create_range("description"), f"'{self.phx_polygon.display_name}"
            ),
            XLItemWindows(create_range("host"), self.phpp_host_surface_id_name),
            XLItemWindows(create_range("glazing_id"), self.phpp_id_glazing),
            XLItemWindows(create_range("frame_id"), self.phpp_id_frame),
            XLItemWindows(
                create_range("width"),
                self.phx_polygon.width,
                "M",
                self._get_target_unit("width"),
            ),
            XLItemWindows(
                create_range("height"),
                self.phx_polygon.height,
                "M",
                self._get_target_unit("height"),
            ),
            # -- TODO: Install condition, not Psi-Install
            XLItemWindows(
                create_range("psi_i_left"), self.phx_construction.frame_left.psi_install
            ),
            XLItemWindows(
                create_range("psi_i_right"),
                self.phx_construction.frame_right.psi_install,
            ),
            XLItemWindows(
                create_range("psi_i_bottom"),
                self.phx_construction.frame_bottom.psi_install,
            ),
            XLItemWindows(
                create_range("psi_i_top"), self.phx_construction.frame_top.psi_install
            ),
        ]

        items_merged = xl_data.merge_xl_item_row(items)

        return items_merged



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller for managing the PHPP Connection."""

from typing import List, Dict

from PHX.model import project, certification, building, components
from PHX.model.hvac.collection import NoVentUnitFoundError

from PHX.xl import xl_app
from PHX.xl.xl_typing import xl_Sheet_Protocol
from PHX.PHPP import sheet_io, phpp_localization
from PHX.PHPP.phpp_localization.shape_model import PhppShape
from PHX.PHPP.phpp_model import (
    areas_surface,
    areas_data,
    areas_thermal_bridges,
    climate_entry,
    electricity_item,
    uvalues_constructor,
    component_glazing,
    component_frame,
    component_vent,
    ventilation_data,
    windows_rows,
    shading_rows,
    vent_space,
    vent_units,
    vent_ducts,
    verification_data,
    hot_water_tank,
    hot_water_piping,
    version,
)


class PHPPConnection:
    """Interface for a PHPP Excel Document."""

    def __init__(self, _xl: xl_app.XLConnection):
        # -- Setup the Excel connection and facade object.
        self.xl = _xl

        # -- Get the localized (units, language) PHPP Shape with worksheet names and column locations
        self.version = self.get_phpp_version()
        self.shape: PhppShape = phpp_localization.get_phpp_shape(self.xl, self.version)

        # -- Setup all the individual worksheet Classes.
        self.verification = sheet_io.Verification(self.xl, self.shape.VERIFICATION)
        self.climate = sheet_io.Climate(self.xl, self.shape.CLIMATE)
        self.u_values = sheet_io.UValues(self.xl, self.shape.UVALUES)
        self.components = sheet_io.Components(self.xl, self.shape.COMPONENTS)
        self.areas = sheet_io.Areas(self.xl, self.shape.AREAS)
        self.windows = sheet_io.Windows(self.xl, self.shape.WINDOWS)
        self.shading = sheet_io.Shading(self.xl, self.shape.SHADING)
        self.addnl_vent = sheet_io.AddnlVent(self.xl, self.shape.ADDNL_VENT)
        self.heating = sheet_io.HeatingDemand(self.xl, self.shape.HEATING_DEMAND)
        self.heating_load = sheet_io.HeatingPeakLoad(
            self.xl, self.shape.HEATING_PEAK_LOAD
        )
        self.cooling = sheet_io.CoolingDemand(self.xl, self.shape.COOLING_DEMAND)
        self.cooling_load = sheet_io.CoolingPeakLoad(
            self.xl, self.shape.COOLING_PEAK_LOAD
        )
        self.ventilation = sheet_io.Ventilation(self.xl, self.shape.VENTILATION)
        self.hot_water = sheet_io.HotWater(self.xl, self.shape.DHW)
        self.electricity = sheet_io.Electricity(self.xl, self.shape.ELECTRICITY)
        self.variants = sheet_io.Variants(self.xl, self.shape.VARIANTS)
        self.per = sheet_io.PER(self.xl, self.shape.PER)
        self.overview = sheet_io.Overview(self.xl, self.shape.OVERVIEW)
        self.use_non_res = sheet_io.UseNonRes(self.xl, self.shape.USE_NON_RES)
        self.elec_non_res = sheet_io.ElecNonRes(self.xl, self.shape.ELEC_NON_RES)
        self.ihg_non_res = sheet_io.IhgNonRes(self.xl, self.shape.IHG_NON_RES)
        self.cooling_units = sheet_io.CoolingUnits(self.xl, self.shape.COOLING_UNITS)
        self.solar_dhw = sheet_io.SolarDHW(self.xl, self.shape.SOLAR_DHW)
        self.solar_pv = sheet_io.SolarPV(self.xl, self.shape.SOLAR_PV)

    def get_data_worksheet(self) -> xl_Sheet_Protocol:
        """Return the 'Data' worksheet from the active PHPP file, support English, German, Spanish."""
        valid_data_worksheet_names = ["DATA", "DATEN", "DATOS"]
        worksheet_names = self.xl.get_worksheet_names()
        for worksheet_name in valid_data_worksheet_names:
            if worksheet_name in worksheet_names:
                return self.xl.get_sheet_by_name(worksheet_name)

        raise Exception(
            f"Error: Cannot find a '{valid_data_worksheet_names}' worksheet in the Excel file: {self.xl.wb.name}?"
        )

    def get_phpp_version(
        self,
        _search_col: str = "A",
        _row_start: int = 1,
        _row_end: int = 10,
    ) -> version.PHPPVersion:
        """Find the PHPP Version and Language of the active xl-file.

        Arguments:
        ----------
            * _xl (xl_app.XLConnection):
            * _search_col (str)
            * _row_start (int) default=1
            * _row_end (int) default=10

        Returns:
        --------
            * PHPPVersion: The PHPPVersion with a number and Language for the Active PHPP.
        """

        # ---------------------------------------------------------------------
        # -- Find the right 'Data' worksheet
        data_worksheet: xl_Sheet_Protocol = self.get_data_worksheet()

        # ---------------------------------------------------------------------
        # -- Pull the search Column data from the Active XL Instance
        data = self.xl.get_single_column_data(
            _sheet_name=data_worksheet.name,
            _col=_search_col,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, xl_rang_data in enumerate(data, start=_row_start):
            if str(xl_rang_data).upper().strip().replace(" ", "").startswith("PHPP"):
                data_row = i
                break
        else:
            raise Exception(
                f"Error: Cannot determine the PHPP Version? Expected 'PHPP' in"
                f"col: {_search_col} of the {data_worksheet.name} worksheet?"
            )

        # ---------------------------------------------------------------------
        # -- Pull the search row data from the Active XL Instance
        data = self.xl.get_single_row_data(data_worksheet.name, data_row)
        data = [_ for _ in data if _ is not None and _ != ""]  # Filter out all the blanks

        # ---------------------------------------------------------------------
        # -- Find the right Version number
        raw_version_id: str = str(
            data[1]
        )  # Use the second value in the row - data (will this always work?)
        ver_major, ver_minor = raw_version_id.split(".")

        # ---------------------------------------------------------------------
        # - Figure out the PHPP language
        # - In <v10 the actual language is not noted in the 'Data' worksheet, so
        # - use the PE factor name as a proxy
        language_search_data = {
            "1-PE-FAKTOREN": "DE",
            "1-FACTORES EP": "ES",
            "1-PE-FACTORS": "EN",
        }
        language = None
        for search_string in language_search_data.keys():
            if search_string in str(data[-1]).upper().strip():
                language = language_search_data[search_string]
                language = language.strip().replace(" ", "_").replace(".", "_")
                break
        if not language:
            raise Exception(
                "Error: Cannot determine the PHPP language? Only English, German and Spanish are supported."
            )

        # ---------------------------------------------------------------------
        # -- Build the new PHPPVersion object
        return version.PHPPVersion(ver_major, ver_minor, language)

    def phpp_version_equals_phx_phi_cert_version(
        self, _phx_variant: project.PhxVariant
    ) -> bool:
        """Return True if the PHX PHI Certification Version and the PHPP Version match."""
        if not int(_phx_variant.phi_certification_major_version) == int(
            self.version.number_major
        ):
            return False
        return True

    def write_certification_config(self, phx_project: project.PhxProject) -> None:
        for phx_variant in phx_project.variants:
            # TODO: how to handle multiple variants?

            if not self.phpp_version_equals_phx_phi_cert_version(phx_variant):
                # -- If the versions don't match, don't try and write anything.
                msg = (
                    f"\nPHPPVersionWarning: the HBJSON PHI "
                    f"Certification version (V={phx_variant.phi_certification_major_version}) "
                    f"does not match the PHPP Version (V={self.version.number_major})? "
                    f"Ignoring all writes to the '{self.shape.VERIFICATION.name}' worksheet.\n"
                )
                self.xl.output(msg)
                return

            # --- Building Type / Use
            self.verification.write_item(
                verification_data.VerificationInput.enum(
                    shape=self.shape.VERIFICATION,
                    input_type="phi_building_category_type",
                    input_enum_value=phx_variant.phi_cert.phi_certification_settings.phi_building_category_type,
                )
            )
            self.verification.write_item(
                verification_data.VerificationInput.enum(
                    shape=self.shape.VERIFICATION,
                    input_type="phi_building_use_type",
                    input_enum_value=phx_variant.phi_cert.phi_certification_settings.phi_building_use_type,
                )
            )
            self.verification.write_item(
                verification_data.VerificationInput.enum(
                    shape=self.shape.VERIFICATION,
                    input_type="phi_building_ihg_type",
                    input_enum_value=phx_variant.phi_cert.phi_certification_settings.phi_building_ihg_type,
                )
            )
            self.verification.write_item(
                verification_data.VerificationInput.enum(
                    shape=self.shape.VERIFICATION,
                    input_type="phi_building_occupancy_type",
                    input_enum_value=phx_variant.phi_cert.phi_certification_settings.phi_building_occupancy_type,
                )
            )

            # --- Certification Config
            self.verification.write_item(
                verification_data.VerificationInput.enum(
                    shape=self.shape.VERIFICATION,
                    input_type="phi_certification_type",
                    input_enum_value=phx_variant.phi_cert.phi_certification_settings.phi_certification_type,
                )
            )
            self.verification.write_item(
                verification_data.VerificationInput.enum(
                    shape=self.shape.VERIFICATION,
                    input_type="phi_certification_class",
                    input_enum_value=phx_variant.phi_cert.phi_certification_settings.phi_certification_class,
                )
            )
            self.verification.write_item(
                verification_data.VerificationInput.enum(
                    shape=self.shape.VERIFICATION,
                    input_type="phi_pe_type",
                    input_enum_value=phx_variant.phi_cert.phi_certification_settings.phi_pe_type,
                )
            )
            self.verification.write_item(
                verification_data.VerificationInput.enum(
                    shape=self.shape.VERIFICATION,
                    input_type="phi_enerphit_type",
                    input_enum_value=phx_variant.phi_cert.phi_certification_settings.phi_enerphit_type,
                )
            )
            self.verification.write_item(
                verification_data.VerificationInput.enum(
                    shape=self.shape.VERIFICATION,
                    input_type="phi_retrofit_type",
                    input_enum_value=phx_variant.phi_cert.phi_certification_settings.phi_retrofit_type,
                )
            )

            # ---- Model Parameters
            if not phx_variant.phius_cert.ph_building_data:
                continue
            self.verification.write_item(
                verification_data.VerificationInput.item(
                    shape=self.shape.VERIFICATION,
                    input_type="num_of_units",
                    input_data=phx_variant.phius_cert.ph_building_data.num_of_units,
                )
            )
            self.verification.write_item(
                verification_data.VerificationInput.item(
                    shape=self.shape.VERIFICATION,
                    input_type="setpoint_winter",
                    input_data=phx_variant.phius_cert.ph_building_data.setpoints.winter,
                    input_unit="C",
                    target_unit=self.shape.VERIFICATION.setpoint_winter.unit,
                )
            )
            self.verification.write_item(
                verification_data.VerificationInput.item(
                    shape=self.shape.VERIFICATION,
                    input_type="setpoint_summer",
                    input_data=phx_variant.phius_cert.ph_building_data.setpoints.summer,
                    input_unit="C",
                    target_unit=self.shape.VERIFICATION.setpoint_summer.unit,
                )
            )

        return None

    def write_climate_data(self, phx_project: project.PhxProject) -> None:
        """Write the variant's weather-station data to the PHPP 'Climate' worksheet."""

        for phx_variant in phx_project.variants:
            # -- Write the actual weather station data
            weather_station_data = climate_entry.ClimateDataBlock(
                shape=self.shape.CLIMATE, phx_site=phx_variant.site
            )
            self.climate.write_climate_block(weather_station_data)

            # -- Set the active weather station
            active_climate_data = climate_entry.ClimateSettings(
                shape=self.shape.CLIMATE, phx_site=phx_variant.site
            )
            self.climate.write_active_climate(active_climate_data)
        return None

    def write_project_constructions(self, phx_project: project.PhxProject) -> None:
        """Write all of the opaque constructions to the PHPP 'U-Values' worksheet."""

        construction_blocks: List[uvalues_constructor.ConstructorBlock] = []
        for phx_construction in phx_project.assembly_types.values():
            construction_blocks.append(
                uvalues_constructor.ConstructorBlock(
                    shape=self.shape.UVALUES, phx_construction=phx_construction
                )
            )

        self.u_values.write_constructor_blocks(construction_blocks)
        return None

    def write_project_window_components(self, phx_project: project.PhxProject) -> None:
        """Write all of the frame and glass constructions from a PhxProject to the PHPP 'Components' worksheet."""

        glazing_component_rows: List[component_glazing.GlazingRow] = []
        frame_component_rows: List[component_frame.FrameRow] = []
        for phx_construction in phx_project.window_types.values():
            glazing_component_rows.append(
                component_glazing.GlazingRow(
                    shape=self.shape.COMPONENTS, phx_construction=phx_construction
                )
            )
            frame_component_rows.append(
                component_frame.FrameRow(
                    shape=self.shape.COMPONENTS, phx_construction=phx_construction
                )
            )
        self.components.write_glazings(glazing_component_rows)
        self.components.write_frames(frame_component_rows)
        return None

    def write_project_ventilation_components(
        self, phx_project: project.PhxProject
    ) -> None:
        """Write all of the ventilators from a PhxProject to the PHPP 'Components' worksheet."""

        phpp_ventilator_rows: List[component_vent.VentilatorRow] = []
        for phx_variant in phx_project.variants:
            for phx_ventilator in phx_variant.mech_systems.ventilation_devices:
                new_vent_row = component_vent.VentilatorRow(
                    shape=self.shape.COMPONENTS,
                    phx_vent_sys=phx_ventilator,
                )
                phpp_ventilator_rows.append(new_vent_row)
        self.components.write_ventilators(phpp_ventilator_rows)
        return None

    def write_project_tfa(self, phx_project: project.PhxProject) -> None:
        for phx_variant in phx_project.variants:
            self.areas.write_item(
                areas_data.AreasInput(
                    shape=self.shape.AREAS,
                    input_type="tfa_input",
                    input_data=phx_variant.building.weighted_net_floor_area,
                    input_unit="M2",
                    target_unit=self.shape.AREAS.tfa_input.unit,
                )
            )
        return None

    def write_project_opaque_surfaces(self, phx_project: project.PhxProject) -> None:
        """Write all of the opaque surfaces from a PhxProject to the PHPP 'Areas' worksheet."""

        surfaces: List[areas_surface.SurfaceRow] = []
        for phx_variant in phx_project.variants:
            for opaque_component in phx_variant.building.opaque_components:
                for phx_polygon in opaque_component.polygons:
                    surfaces.append(
                        areas_surface.SurfaceRow(
                            self.shape.AREAS,
                            phx_polygon,
                            opaque_component,
                            self.u_values.get_constructor_phpp_id_by_name(
                                opaque_component.assembly.display_name, _use_cache=True
                            ),
                        )
                    )

        if len(surfaces) >= 100:
            print(
                f"Warning: {len(surfaces)} surfaces found in the model. Ensure that you have "
                "added enough rows to the 'Areas' worksheet to handle that many surfaces. "
                "By default the PHPP can only have 100 surfaces input."
            )

        self.areas.write_surfaces(surfaces)
        return None

    def write_project_thermal_bridges(self, phx_project: project.PhxProject) -> None:
        """Write all of the thermal-bridge elements of a PhxProject to the PHPP 'Areas' worksheet."""

        thermal_bridges: List[areas_thermal_bridges.ThermalBridgeRow] = []
        for variant in phx_project.variants:
            for zone in variant.zones:
                for phx_tb in zone.thermal_bridges:
                    thermal_bridges.append(
                        areas_thermal_bridges.ThermalBridgeRow(self.shape.AREAS, phx_tb)
                    )

        if len(thermal_bridges) >= 100:
            print(
                f"Warning: {len(thermal_bridges)} thermal bridges found in the model. Ensure that you have "
                "added enough rows to the 'Areas' worksheet to handle that many thermal bridges. "
                "By default the PHPP can only have 100 thermal bridges input."
            )

        self.areas.write_thermal_bridges(thermal_bridges)
        return None

    def write_project_window_surfaces(self, phx_project: project.PhxProject) -> None:
        """Write all of the window surfaces from a PhxProject to the PHPP 'Windows' worksheet."""

        # -- First, populate the Variant types
        variant_type_names = [
            phx_aperture.variant_type_name
            for phx_variant in phx_project.variants
            for phx_component in phx_variant.building.opaque_components
            for phx_aperture in phx_component.apertures
        ]
        for i, variant_type_name in enumerate(sorted(variant_type_names)):
            self.variants.write_window_type(variant_type_name, i)

        # -- Get the variant types in a dict
        window_type_phpp_ids = self.variants.get_window_type_phpp_ids()

        # -- Write in the window-data
        phpp_windows: List[windows_rows.WindowRow] = []
        for phx_variant in phx_project.variants:
            for phx_component in phx_variant.building.opaque_components:
                for phx_aperture in phx_component.apertures:
                    for ap_polygon in phx_aperture.polygons:
                        host_polygon = phx_component.get_host_polygon_by_child_id_num(
                            ap_polygon.id_num
                        )
                        phpp_host_surface_id_name = (
                            self.areas.surfaces.get_surface_phpp_id_by_name(
                                host_polygon.display_name, _use_cache=True
                            )
                        )
                        phpp_id_frame = self.components.frames.get_frame_phpp_id_by_name(
                            phx_aperture.window_type.frame_type_display_name,
                            _use_cache=True,
                        )
                        phpp_id_glazing = (
                            self.components.glazings.get_glazing_phpp_id_by_name(
                                phx_aperture.window_type.glazing_type_display_name,
                                _use_cache=True,
                            )
                        )

                        phpp_windows.append(
                            windows_rows.WindowRow(
                                shape=self.shape.WINDOWS,
                                phx_polygon=ap_polygon,
                                phx_construction=phx_aperture.window_type,
                                phpp_host_surface_id_name=phpp_host_surface_id_name,
                                phpp_id_frame=phpp_id_frame,
                                phpp_id_glazing=phpp_id_glazing,
                                phpp_id_variant_type=window_type_phpp_ids[
                                    phx_aperture.variant_type_name
                                ].phpp_id,
                            )
                        )

        if len(phpp_windows) >= 150:
            print(
                f"Warning: {len(phpp_windows)} windows found in the model. Ensure that you have "
                "added enough rows to the 'Windows' worksheet to handle that many windows. "
                "By default the PHPP can only have 150 windows input."
            )

        self.windows.write_windows(phpp_windows)
        return None

    def write_project_window_shading(self, phx_project: project.PhxProject) -> None:
        def _get_ap_element_from_dict(
            _window_name: str, _dict: Dict[str, components.PhxApertureElement]
        ) -> components.PhxApertureElement:
            """When reading from excel, it MIGHT come in as a float. This means
            that any windows with a numerical name (ie: '106', '205', etc) will get
            converted to a float in excel (ie: '106' -> 106.0) and get read back in
            that way. To support names 106, 106.0, '106' and '106.0' properly, try them
            with a fallback sequence.
            """

            try:
                # If its a normal string name...
                return _dict[_window_name]
            except KeyError:
                try:
                    # If it is a 'float' name (ie: 106.1)
                    return _dict[str(float(_window_name))]
                except KeyError:
                    try:
                        # If it is an 'int' name (ie: 106)
                        return _dict[str(int(float(_window_name)))]
                    except KeyError as e:
                        raise e

        # Get all the Window worksheet names in order
        window_names = self.windows.get_all_window_names()

        # Get all the PHX Aperture objects
        phx_aperture_dict: Dict[str, components.PhxApertureElement] = {}
        for phx_variant in phx_project.variants:
            for phx_component in phx_variant.building.opaque_components:
                for phx_aperture in phx_component.apertures:
                    for phx_ap_element in phx_aperture.elements:
                        phx_aperture_dict[phx_ap_element.display_name] = phx_ap_element

        # Sort the phx apertures to match the window_names order
        phx_aperture_elements_in_order = (
            _get_ap_element_from_dict(window_name, phx_aperture_dict)
            for window_name in window_names
        )

        # Write out all the data to the Shading Worksheet
        phpp_shading_rows: List[shading_rows.ShadingRow] = []
        for phx_aperture_element in phx_aperture_elements_in_order:
            phpp_shading_rows.append(
                shading_rows.ShadingRow(
                    self.shape.SHADING,
                    phx_aperture_element.shading_dimensions,
                    phx_aperture_element.winter_shading_factor,
                    phx_aperture_element.summer_shading_factor,
                )
            )
        self.shading.write_shading(phpp_shading_rows)

        # TODO: option to clear all the dimensional info?

        return None

    def write_project_ventilators(self, phx_project: project.PhxProject) -> None:
        """Write all of the used Ventilator Units from a PhxProject to the PHPP 'Additional Vent' worksheet."""

        phpp_vent_unit_rows: List[vent_units.VentUnitRow] = []
        for phx_variant in phx_project.variants:
            for phx_ventilator in phx_variant.mech_systems.ventilation_devices:
                phpp_id_ventilator = (
                    self.components.ventilators.get_ventilator_phpp_id_by_name(
                        phx_ventilator.display_name
                    )
                )
                new_vent_row = vent_units.VentUnitRow(
                    shape=self.shape.ADDNL_VENT,
                    phx_vent_sys=phx_ventilator,
                    phpp_id_ventilator=phpp_id_ventilator,
                )
                phpp_vent_unit_rows.append(new_vent_row)

        self.addnl_vent.write_vent_units(phpp_vent_unit_rows)
        return None

    def write_project_spaces(self, phx_project: project.PhxProject) -> None:
        """Write all of the PH-Spaces from a PhxProject to the PHPP 'Additional Vent' worksheet."""

        phpp_vent_rooms: List[vent_space.VentSpaceRow] = []
        for phx_variant in phx_project.variants:
            for zone in phx_variant.building.zones:
                for room in zone.spaces:
                    # -- Find the right Ventilator assigned to the Space.
                    try:
                        phx_mech_ventilator = (
                            phx_variant.mech_systems.get_mech_device_by_id(
                                room.vent_unit_id_num
                            )
                        )
                        phpp_id_ventilator = (
                            self.components.ventilators.get_ventilator_phpp_id_by_name(
                                phx_mech_ventilator.display_name
                            )
                        )
                        phpp_row_ventilator = (
                            self.addnl_vent.vent_units.get_vent_unit_num_by_phpp_id(
                                phpp_id_ventilator
                            )
                        )
                    except NoVentUnitFoundError:
                        # If no ventilation system / unit has not been applied yet
                        phpp_row_ventilator = None

                    phx_vent_pattern = phx_project.utilization_patterns_ventilation.get_pattern_by_id_num(
                        room.ventilation.schedule.id_num
                    )

                    phpp_rm = vent_space.VentSpaceRow(
                        shape=self.shape.ADDNL_VENT,
                        phx_room_vent=room,
                        phpp_row_ventilator=phpp_row_ventilator,
                        phx_vent_pattern=phx_vent_pattern,
                    )
                    phpp_vent_rooms.append(phpp_rm)

        if len(phpp_vent_rooms) >= 30:
            print(
                f"Warning: {len(phpp_vent_rooms)} spaces found in the model. Ensure that you have "
                "added enough rows to the 'Additional Vent' worksheet to handle that many spaces. "
                "By default the PHPP can only have 30 spaces input."
            )

        self.addnl_vent.write_spaces(phpp_vent_rooms)
        return None

    def write_project_ventilation_type(self, phx_project: project.PhxProject) -> None:
        """Set the Ventilation-Type to the PHPP 'Ventilation' worksheet."""

        for variant in phx_project.variants:
            self.ventilation.write_ventilation_type(
                # TODO: Get the actual type from the model someplace?
                # TODO: How to combine Variants?
                ventilation_data.VentilationInputItem.vent_type(
                    self.shape.VENTILATION, "1-Balanced PH ventilation with HR"
                )
            )
            self.ventilation.write_multi_vent_worksheet_on(
                ventilation_data.VentilationInputItem.multi_unit_on(
                    self.shape.VENTILATION, "x"
                )
            )
        return None

    def write_project_volume(self, phx_project: project.PhxProject) -> None:
        """Write the Vn50 and Vv to the PHPP 'Ventilation Worksheet."""
        for variant in phx_project.variants:
            # TODO: How to handle multiple variants?

            if not variant.phius_cert.ph_building_data:
                continue
            bldg: building.PhxBuilding = variant.building

            self.ventilation.write_Vn50_volume(
                ventilation_data.VentilationInputItem.airtightness_Vn50(
                    self.shape.VENTILATION, bldg.net_volume
                )
            )

    def write_project_airtightness(self, phx_project: project.PhxProject) -> None:
        """Write the Airtightness data to the PHPP 'Ventilation' worksheet."""

        for variant in phx_project.variants:
            # TODO: How to handle multiple variants?

            if not variant.phius_cert.ph_building_data:
                continue
            ph_bldg: certification.PhxPhBuildingData = variant.phius_cert.ph_building_data

            # TODO: Get the actual values from the Model somehow
            self.ventilation.write_wind_coeff_e(
                ventilation_data.VentilationInputItem.wind_coeff_e(
                    self.shape.VENTILATION, ph_bldg.wind_coefficient_e
                )
            )
            self.ventilation.write_wind_coeff_f(
                ventilation_data.VentilationInputItem.wind_coeff_f(
                    self.shape.VENTILATION, ph_bldg.wind_coefficient_f
                )
            )
            self.ventilation.write_airtightness_n50(
                ventilation_data.VentilationInputItem.airtightness_n50(
                    self.shape.VENTILATION, ph_bldg.airtightness_n50
                )
            )
        return None

    def write_project_hot_water(self, phx_project: project.PhxProject) -> None:
        """Write the Hot Water data to the PHPP 'DHW+Distribution' worksheet."""
        for variant in phx_project.variants:
            # -- Tanks
            # Use only the first 2 tanks for PHPP
            if len(variant.mech_systems.dhw_tank_devices) > 2:
                print(
                    f"Warning: PHPP only allows 2 tanks."
                    f"{len(variant.mech_systems.dhw_tank_devices)} tank"
                    f'found in the Variant "{variant.name}"'
                )

            tank_inputs = []
            for i, phx_dhw_tank in enumerate(
                variant.mech_systems.dhw_tank_devices[:2], start=1
            ):
                tank_inputs.append(
                    hot_water_tank.TankInput(
                        self.shape.DHW,
                        phx_dhw_tank,  # type: water.PhxHotWaterTank
                        i,
                    )
                )
            self.hot_water.write_tanks(tank_inputs)

            # -- Branch Piping
            branch_piping_inputs = []
            branch_pipe_groups = variant.mech_systems.dhw_branch_piping_segments_by_diam
            if len(branch_pipe_groups) > 5:
                print(
                    "Warning: PHPP only allows 5 groups of DHW branch piping. "
                    f"{len(branch_pipe_groups)} piping groups "
                    f'found in the Variant "{variant.name}". '
                    "Using only then first 5 piping groups."
                )

            for i, phx_branch_piping in enumerate(branch_pipe_groups[:5]):
                branch_piping_inputs.append(
                    hot_water_piping.BranchPipingInput(
                        self.shape.DHW,
                        phx_branch_piping,  # type: piping.PhxPipeSegment
                        i,
                        variant.mech_systems._distribution_num_hw_tap_points,
                    )
                )
            self.hot_water.write_branch_piping(branch_piping_inputs)

            # -- Recirculation Piping
            recirc_piping_inputs = []
            recirc_pipe_groups = variant.mech_systems.dhw_recirc_piping_segments_by_diam
            if len(recirc_pipe_groups) > 5:
                print(
                    "Warning: PHPP only allows 5 groups of DHW Recirc. piping. "
                    f"{len(recirc_pipe_groups)} piping groups "
                    f'found in the Variant "{variant.name}". '
                    "Using only then first 5 piping groups."
                )

            for i, phx_recirc_piping in enumerate(recirc_pipe_groups[:5]):
                recirc_piping_inputs.append(
                    hot_water_piping.RecircPipingInput(
                        self.shape.DHW,
                        phx_recirc_piping,  # type: piping.PhxPipeSegment
                        i,
                    )
                )
            self.hot_water.write_recirc_piping(recirc_piping_inputs)

        return None

    def write_project_res_elec_appliances(self, phx_project: project.PhxProject) -> None:
        """Write out all of the detailed residential appliances to the "Electricity" Worksheet."""
        equipment_inputs = []
        for phx_variant in phx_project.variants:
            for zone in phx_variant.building.zones:
                for phx_equip in zone.elec_equipment_collection:
                    equipment_inputs.append(
                        electricity_item.ElectricityItemXLWriter(phx_equip)
                    )
            self.electricity.write_equipment(equipment_inputs)

        return None

    def write_non_res_utilization_profiles(self, phx_project: project.PhxProject) -> None:
        """Write out all of the Utilization patterns to the "Use non-res" Worksheet."""
        # TODO: build this....
        # for pattern in phx_project.utilization_patterns_occupancy:
        #     print(pattern)

    def write_non_res_space_lighting(self, phx_project: project.PhxProject) -> None:
        """Write out all of the Space Lighting values to the "Electricity non-res" Worksheet."""
        return

    def write_non_res_IHG(self, phx_project: project.PhxProject) -> None:
        """Write out all of the Occupancy patterns to the "IHG non-res" Worksheet."""
        return

    def activate_variant_assemblies(self) -> None:
        """Remove all existing U-Value information and link assemblies to the Variants worksheet."""

        # -- Collect all the assemblies from the U-Values page
        # -- and add each one to the Variants assembly-layers section
        for i, assembly_name in enumerate(
            self.u_values.get_used_constructor_names(), start=0
        ):
            if i > 25:
                print(
                    "WARNING: The Variants worksheet can only handle 26 different assemblies."
                    "You will have to set up the assembly-layer Variants links manually."
                )
                continue
            self.variants.write_assembly_layer(assembly_name, i)

        # -- Get all the Variant assembly names with prefix
        assembly_phpp_ids = self.variants.get_assembly_layer_phpp_ids()

        # -- Make all Variants--->U-Values links
        self.u_values.activate_variants(assembly_phpp_ids)

        return None

    def activate_variant_windows(self) -> None:
        """Set the Frame and Glass Components to link to the Variants worksheet for all windows."""

        self.windows.activate_variants()

        return None

    def activate_variant_ventilation(self) -> None:
        """Set the ACH, Ventilation type to link to the Variants worksheet."""

        self.ventilation.activate_variants()

        return None

    def activate_variant_additional_vent(self) -> None:
        """Set the Ventilator, duct length and insulation to link to the Variants worksheet."""

        # -- Find the locations of the Ventilation input items
        input_item_rows = self.variants.get_ventilation_input_item_rows()
        vent_unit_row = input_item_rows[
            self.shape.VARIANTS.ventilation.input_item_names.ventilator_unit
        ]

        self.addnl_vent.activate_variants(
            variants_worksheet_name=self.shape.VARIANTS.name,
            vent_unit_range=f"{self.shape.VARIANTS.active_value_column}{vent_unit_row}",
        )

        return None



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions used to create Project elements from the Honeybee-Model"""

from typing import Union, Optional, Any, Tuple, List

from honeybee import model
from honeybee.aperture import Aperture
from honeybee_energy.material.opaque import EnergyMaterial, EnergyMaterialNoMass
from honeybee_energy.construction import window, windowshade

from honeybee_energy.construction.window import WindowConstruction
from honeybee_energy.construction.windowshade import WindowConstructionShade
from honeybee_energy_ph.properties.construction.window import (
    WindowConstructionPhProperties,
)
from honeybee_energy_ph.construction.window import PhWindowGlazing, PhWindowFrame
from honeybee_ph_utils import iso_10077_1

from PHX.model import constructions, project, shades


def _conductivity_from_r_value(_r_value: float, _thickness: float) -> float:
    """Returns a material conductivity value (W/mk), given a known r-value (M2K/W) and thickness (M).

    Arguments:
    ----------
        * _r_value (float): The total R-Value of the layer.
        * _thickness (float): The total thickness of the layer.

    Returns:
    --------
        * float: The Conductivity value of the layer.
    """
    conductivity = _thickness / _r_value
    return conductivity


def build_phx_material_from_hb_EnergyMaterial(
    _hb_material: EnergyMaterial,
) -> constructions.PhxMaterial:
    """Returns a new PhxMaterial, with attributes based on a Honeybee-Energy EnergyMaterial.

    Arguments:
    ----------
        * _hb_material (EnergyMaterialNoMass): The Honeybee-Energy EnergyMaterialNoMass
        * _thickness (float): The thickness of the material layer.

    Returns:
    --------
        * (constructions.PhxMaterial): The new PhxMaterial with attributes based on the
            Honeybee-Energy EnergyMaterial.
    """
    new_mat = constructions.PhxMaterial()
    new_mat.display_name = _hb_material.display_name
    new_mat.conductivity = _hb_material.conductivity
    new_mat.density = _hb_material.density
    new_mat.heat_capacity = _hb_material.specific_heat
    new_mat.percentage_of_assembly = _hb_material.properties.ph.percentage_of_assembly  # type: ignore

    # -- Defaults
    new_mat.porosity = 0.95
    new_mat.water_vapor_resistance = 1.0
    new_mat.reference_water = 0.0

    return new_mat


def build_phx_material_from_hb_EnergyMaterialNoMass(
    _hb_material: EnergyMaterialNoMass, _thickness: float
) -> constructions.PhxMaterial:
    """Returns a new PhxMaterial, with attributes based on a Honeybee-Energy EnergyMaterialNoMass.
    Note that a thickness is required in order to determine the material conductivity.

    Arguments:
    ----------
        * _hb_material (EnergyMaterialNoMass): The Honeybee-Energy EnergyMaterialNoMass
        * _thickness (float): The thickness of the material layer.

    Returns:
    --------
        * (constructions.PhxMaterial): The new PhxMaterial with attributes based on the
            Honeybee-Energy EnergyMaterialNoMass.
    """
    new_mat = constructions.PhxMaterial()
    new_mat.display_name = _hb_material.display_name
    new_mat.conductivity = _conductivity_from_r_value(_hb_material.r_value, _thickness)
    new_mat.density = _hb_material.mass_area_density
    new_mat.heat_capacity = _hb_material.area_heat_capacity

    # -- Defaults
    new_mat.water_vapor_resistance = 1.0
    new_mat.porosity = 0.95
    new_mat.reference_water = 0.0

    return new_mat


def build_layer_from_hb_material(
    _hb_material: Union[EnergyMaterial, EnergyMaterialNoMass]
) -> constructions.PhxLayer:
    """Returns a new PHX-Layer with attributes based on a Honeybee-Material.

    Arguments:
    ----------
        *_hb_material (EnergyMaterial | EnergyMaterialNoMass): The Honeybee Material.

    Returns:
    --------
        * constructions.Layer: The new PHX-Layer object.
    """
    new_layer = constructions.PhxLayer()

    if isinstance(_hb_material, EnergyMaterial):
        new_layer.thickness_m = _hb_material.thickness

        if _hb_material.properties.ph.base_materials:  # type: ignore
            ## -- Layer is a heterogeneous material layer (ie: stud + insulation)
            for hb_sub_material in _hb_material.properties.ph.base_materials:  # type: ignore
                new_layer.add_material(
                    build_phx_material_from_hb_EnergyMaterial(hb_sub_material)
                )
        else:
            # -- Layer is homogeneous, use the base HB Material
            new_layer.add_material(
                build_phx_material_from_hb_EnergyMaterial(_hb_material)
            )

    elif isinstance(_hb_material, EnergyMaterialNoMass):
        # -- NoMass layers cannot be heterogeneous (yet....)
        new_layer.thickness_m = (
            0.1  # 0.1m = 4". Use as default since No-Mass has no thickness
        )
        new_layer.add_material(
            build_phx_material_from_hb_EnergyMaterialNoMass(
                _hb_material, new_layer.thickness_m
            )
        )

    else:
        raise TypeError(f"Error: Unsupported Material type: '{type(_hb_material)}'.")

    return new_layer


def build_opaque_assemblies_from_HB_model(
    _project: project.PhxProject, _hb_model: model.Model
) -> None:
    """Build Opaque Constructions from Honeybee Faces and add to the PHX-Project.

    Will also align the id_nums of the face's Construction with the Assembly in the Project dict.

    Arguments:
    ----------
        * _hb_model (model.Model): The Honeybee Model to use as the source.

    Returns:
    --------
        * None
    """

    for room in _hb_model.rooms:
        for face in room.faces:
            hb_const = face.properties.energy.construction

            if not hb_const.identifier in _project.assembly_types:
                # -- Create a new Assembly with Layers from the Honeybee-Construction
                new_assembly = constructions.PhxConstructionOpaque()
                new_assembly.id_num = constructions.PhxConstructionOpaque._count
                new_assembly.display_name = hb_const.display_name
                new_assembly.layers = [
                    build_layer_from_hb_material(layer) for layer in hb_const.materials
                ]

                # -- Add the assembly to the Project
                _project.add_assembly_type(new_assembly, hb_const.identifier)

            # -- Keep the ID numbers in sync
            _phx_assembly = _project.assembly_types[hb_const.identifier]
            hb_const.properties.ph.id_num = _phx_assembly.id_num

    return None


def _set_phx_window_type_glazing(
    _phx_window_type: constructions.PhxConstructionWindow,
    _hbph_glazing: Optional[PhWindowGlazing],
) -> constructions.PhxConstructionWindow:
    if _hbph_glazing:
        # -- Use Detailed PH-Params
        _phx_window_type.glazing_type_display_name = _hbph_glazing.display_name
        _phx_window_type.u_value_glass = _hbph_glazing.u_factor
        _phx_window_type.glass_g_value = _hbph_glazing.g_value
    return _phx_window_type


def _set_phx_window_type_frames(
    _phx_window_type: constructions.PhxConstructionWindow,
    _hbph_frame: Optional[PhWindowFrame],
) -> constructions.PhxConstructionWindow:
    """Sets the Frame properties of a PhxConstructionWindow based on a Honeybee-Ph WindowFrame."""
    if _hbph_frame:
        # -- Use Detailed PH-Params
        _phx_window_type.frame_type_display_name = _hbph_frame.display_name
        _phx_window_type.frame_top.u_value = _hbph_frame.top.u_factor
        _phx_window_type.frame_top.width = _hbph_frame.top.width
        _phx_window_type.frame_top.psi_glazing = _hbph_frame.top.psi_glazing
        _phx_window_type.frame_top.psi_install = _hbph_frame.top.psi_install

        _phx_window_type.frame_right.u_value = _hbph_frame.right.u_factor
        _phx_window_type.frame_right.width = _hbph_frame.right.width
        _phx_window_type.frame_right.psi_glazing = _hbph_frame.right.psi_glazing
        _phx_window_type.frame_right.psi_install = _hbph_frame.right.psi_install

        _phx_window_type.frame_bottom.u_value = _hbph_frame.bottom.u_factor
        _phx_window_type.frame_bottom.width = _hbph_frame.bottom.width
        _phx_window_type.frame_bottom.psi_glazing = _hbph_frame.bottom.psi_glazing
        _phx_window_type.frame_bottom.psi_install = _hbph_frame.bottom.psi_install

        _phx_window_type.frame_left.u_value = _hbph_frame.left.u_factor
        _phx_window_type.frame_left.width = _hbph_frame.left.width
        _phx_window_type.frame_left.psi_glazing = _hbph_frame.left.psi_glazing
        _phx_window_type.frame_left.psi_install = _hbph_frame.left.psi_install
    return _phx_window_type


def _set_phx_window_type_u_w_value(
    _phx_window_type: constructions.PhxConstructionWindow,
    _hbph_frame: Optional[PhWindowFrame],
    _hbph_glazing: Optional[PhWindowGlazing],
) -> constructions.PhxConstructionWindow:
    """Set the U-Value and Frame-Factor of a PhxConstructionWindow based on the given HBPH-Params."""
    if _hbph_frame and _hbph_glazing:
        _phx_window_type.frame_factor = iso_10077_1.calculate_window_frame_factor(
            _hbph_frame, _hbph_glazing
        )
        _phx_window_type.u_value_window = iso_10077_1.calculate_window_uw(
            _hbph_frame, _hbph_glazing
        )
    return _phx_window_type


def build_phx_window_type_from_HB_WindowConstruction(
    _project: project.PhxProject,
    _hb_win_const: window.WindowConstruction,
    _shade_const: Optional[windowshade.WindowConstructionShade],
) -> constructions.PhxConstructionWindow:
    """Create a new PhxConstructionWindow based on a HBPH-WindowConstruction.

    If any detailed PH-Params exist for the frame or glass on the HB-Window-Construction's
    .properties.ph.* then those will be used. Otherwise, the basic HB-Window-Construction
    attributes will be used.

    Arguments:
    ----------
        * _hb_win_const (window.WindowConstruction): The Honeybee Window Construction to
            base the new PHX-WindowType on.

    Returns:
    --------
        * (constructions.WindowType): The new PHX-WindowType.
    """
    phx_window_type = constructions.PhxConstructionWindow()
    phx_window_type.id_num = constructions.PhxConstructionWindow._count
    phx_window_type.display_name = _hb_win_const.display_name
    phx_window_type.identifier = _hb_win_const.identifier

    ph_params: WindowConstructionPhProperties = _hb_win_const.properties.ph  # type: ignore

    # -------------------------------------------------------------------------
    # -- Set the basic data first ---------------------------------------------
    # -- Glass ----------------------------------------------------------------
    phx_window_type.u_value_glass = _hb_win_const.u_factor
    phx_window_type.glass_g_value = _hb_win_const.shgc
    phx_window_type.u_value_window = _hb_win_const.u_factor
    phx_window_type.frame_factor = 0.75

    # -- Frames ---------------------------------------------------------------
    phx_window_type.set_all_frames_u_value(_hb_win_const.u_factor)
    phx_window_type.set_all_frames_width(0.1)
    phx_window_type.set_all_frames_psi_glazing(0.0)
    phx_window_type.set_all_frames_psi_install(0.0)

    # -------------------------------------------------------------------------
    # -- Then set detailed PH data, if any  -----------------------------------
    ph_frame = ph_params.ph_frame
    ph_glazing = ph_params.ph_glazing
    phx_window_type = _set_phx_window_type_glazing(phx_window_type, ph_glazing)
    phx_window_type = _set_phx_window_type_frames(phx_window_type, ph_frame)

    # -- Window Params as per ISO-10077-1 -------------------------------------
    phx_window_type = _set_phx_window_type_u_w_value(
        phx_window_type, ph_frame, ph_glazing
    )

    # -- Add Shading to the Window, if any -------------------------------------
    if _shade_const:
        phx_shade = _project.shade_types[_shade_const.shade_material.identifier]
        phx_window_type._id_num_shade = phx_shade.id_num

    return phx_window_type


def build_phx_shade_type_from_HB_WindowConstructionShade(
    _hb_const: windowshade.WindowConstructionShade,
) -> shades.PhxWindowShade:
    """Create a new PhxWindowShade object with attributes based on an HBE-WindowConstructionShade.

    Arguments:
    ----------
        * _hb_const: (windowshade.WindowConstructionShade) The source HBE-WindowConstructionShade.

    Returns:
    --------
        * (shades.PhxWindowShade): The new PHX Window Shade type.
    """
    _hb_shade_material = _hb_const.shade_material

    phx_shade = shades.PhxWindowShade()
    phx_shade.display_name = _hb_shade_material.display_name
    phx_shade.identifier = _hb_shade_material.identifier
    # TODO: Phius Shade Method?
    phx_shade.reduction_factor = _hb_shade_material.solar_transmittance

    # -- Keep the IDs aligned
    _hb_const.properties.ph.id_num = phx_shade.id_num  # type: ignore

    return phx_shade


def _get_hbph_window_constructions(
    _ap_ep_const,
) -> Tuple[WindowConstruction, Optional[WindowConstructionShade]]:
    """Get the WindowConstruction and WindowConstructionShade from an HB-Aperture Construction."""
    try:
        # -- It is a WindowConstructionShade if it has a 'window_construction' attribute
        hb_win_const = _ap_ep_const.window_construction
        hb_shade_const = _ap_ep_const
    except AttributeError:
        # -- otherwise, its a normal window construction
        hb_win_const = _ap_ep_const
        hb_shade_const = None

    return hb_win_const, hb_shade_const


def _new_shade_type(
    _project: project.PhxProject, hb_shade_const: WindowConstructionShade
) -> bool:
    """Check if a new shade type needs to be created for the given HB-WindowConstructionShade."""
    if hb_shade_const.shade_material.identifier in _project.shade_types:
        return False
    else:
        return True


def build_transparent_assembly_types_from_HB_Model(
    _project: project.PhxProject, _hb_apertures: List[Aperture]
) -> None:
    """Create PHX-WindowTypes (constructions) from an HB Model and add to the PHX-Project

    Will also align the id_nums of the Aperture Construction's with the WindowType in the Project dict.

    Arguments:
    ----------
        * _project (_project: project.PhxProject): The PhxProject to store the new window-type on.
        * _hb_apertures (List[Aperture]): The Honeybee Apertures to use as the source

    Returns:
    --------
        * None
    """

    for aperture in _hb_apertures:
        # ---------------------------------------------------------------------
        ap_ep_const = aperture.properties.energy.construction  # type: ignore
        hb_win_const, hb_shade_const = _get_hbph_window_constructions(ap_ep_const)

        # ---------------------------------------------------------------------
        # --- Build all the PHX Shades first since the window needs to know their ID num
        if hb_shade_const and _new_shade_type(_project, hb_shade_const):
            phx_shade_type = build_phx_shade_type_from_HB_WindowConstructionShade(
                hb_shade_const
            )
            _project.add_new_shade_type(phx_shade_type)

        # ---------------------------------------------------------------------
        # --- Build the PHX Windows
        if hb_win_const.identifier not in _project.window_types:
            phx_aperture_constr = build_phx_window_type_from_HB_WindowConstruction(
                _project, hb_win_const, hb_shade_const
            )
            _project.add_new_window_type(phx_aperture_constr)

        # ---------------------------------------------------------------------
        # -- Keep all the IDs aligned
        phx_win_type = _project.window_types[hb_win_const.identifier]
        hb_win_const.properties.ph.id_num = phx_win_type.id_num  # type: ignore

    return None



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Exceptions used by the IO classes."""


class FindSectionMarkerException(Exception):
    def __init__(self, search_string, _sheet_name, _col_letter):
        """Raises when the IO controller cannot find the reference marker in a column."""
        self.msg = (
            f"\n\tError: Cannot find the the marker: '{search_string}' "
            f"in the worksheet '{_sheet_name}' column '{_col_letter}'?"
        )
        super().__init__(self.msg)


class PerReferenceAreaException(Exception):
    def __init__(self, _sheet_name, _search_address):
        """Raises when the PER reference area (TFA / Footprint) is missing."""
        self.msg = (
            f"\n\tError: Cannot find the the reference area on '{_sheet_name}' "
            f"worksheet at location '{_search_address}'?"
        )
        super().__init__(self.msg)


class ReadDataException(Exception):
    def __init__(self, _sheet_name, _read_address):
        """Raised when there is an error reading a value from Excel."""
        self.msg = (
            f"\n\tError: Cannot read the value from '{_sheet_name}' "
            f"worksheet at location '{_read_address}'?"
        )
        super().__init__(self.msg)


class PHPPDataMissingException(Exception):
    def __init__(self, _sheet_name, _read_address):
        """Raised when there a required field returns 'None'."""
        self.msg = (
            f"Error: Required value missing in PHPP "
            f"worksheet '{_sheet_name}' at cell '{_read_address}'. "
            "Please correct the PHPP."
        )
        super().__init__(self.msg)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Ventilation Load."""

from __future__ import annotations
from typing import ClassVar, Any
from dataclasses import dataclass, field


@dataclass
class PhxLoadVentilation:
    """A PHX Load for the Ventilation."""

    flow_supply: float = 0.0
    flow_extract: float = 0.0
    flow_transfer: float = 0.0
